import weakref
from matplotlib.cm import ScalarMappable
from matplotlib.colors import LogNorm, Normalize, Colormap, SymLogNorm
import ifigure
from ifigure.widgets.undo_redo_history import UndoRedoFigobjMethod
import matplotlib.ticker as mticker
from ifigure.utils.cbook import isiterable
from ifigure.ifigure_config import isMPL33
import ifigure.utils.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('AxisParam')

def _to_float(value):
    if hasattr(value, 'real'):
        return value.real
    return float(value)


def get_cmap(name, param=256):
    from matplotlib import colormaps
    return colormaps[name].resampled(param)

class Memberholder(object):
    def __init__(self):
        self._member = []
        self._ax_idx = []

    def add_ax_idx(self, v):
        if not v in self._ax_idx:
            self._ax_idx.append(v)

    def rm_ax_idx(self, v):
        self._ax_idx = [x for x in self._ax_idx if x != v]

    def get_member(self):
        return self._member

    def walk_member(self):
        for m in self._member:
            if m() is not None:
                yield(m())

    def add_member(self, obj):
        for r in self._member:
            if r() is obj:
                return
        self._member.append(weakref.ref(obj, self._memberdead))

    def num_member(self):
        return len(self._member)

    def rm_member(self, obj):
        for r in self._member:
            if r() is obj:
                self._member.remove(r)
                return

    def _memberdead(self, ref):
        self._member = [a for a in self._member
                        if a() is not None]


class AxisRangeParam(Memberholder):
    def __init__(self):
        Memberholder.__init__(self)
        self.name = ''
        self.base = 10
        self.auto = True
        self.range = [0, 1]
        self.symloglin = 1.0
        self.symloglinscale = 1.0
        self.scale = 'linear'
        self.cmap = 'jet'
        self.ticks = None     # None: auto, number, or exact value
        self.mode = (False, False, False)  # int, #symmetric, #margin

    def make_rangeparam_action(self, artist, value):
        action = UndoRedoFigobjMethod(artist,
                                      'axrangeparam', value)
        action.set_extrainfo(self.name)
        return action

    def make_rangeparam_action0(self, artist, value):
        '''
        this one does call handle_axes_change
        '''
        dprint1('make_rangeparam_action0 is obsolete and should not be used')

        def f(obj=artist.figobj, name=value[0]):
            obj.call_handle_axes_change(name)
        action = UndoRedoFigobjMethod(artist,
                                      'axrangeparam', value,
                                      finish_action=f)
        action.set_extrainfo(self.name)
        return action

    def get_rangeparam(self):
        v = [str(self.base),
             self.auto,
             self.range[:],
             self.scale,
             self.symloglin,
             self.symloglinscale,
             self.mode[0], self.mode[1], self.mode[2]]
        return v

    def set_rangeparam(self, value):
        if len(value) <= 7:
            value = [x for x in value]
            value.append(False)
        self.base = float(value[0])
        self.auto = value[1]
        self.mode = (value[6], value[7], value[8])
        self.symloglin = value[4]
        self.symloglinscale = value[5]
        self.scale = value[3]

        self.range = [_to_float(value[2][0]), _to_float(value[2][1])]

    def set_artist_rangeparam(self,
                              set_scale=None,
                              set_auto=None,
                              set_lim=None,
                              set_ticks=None):

        set_auto(False)
        suffix = self.name
        if self.scale == 'log':
            if isMPL33:
                kargs = {'base': self.base}
            else:
                kargs = {'base' + suffix: self.base}
        elif self.scale == 'symlog':
            if isMPL33:
                kargs = {'base': self.base,
                         'linthresh': self.symloglin,
                         'linscale': self.symloglinscale}
            else:
                kargs = {'base' + suffix: self.base,
                         'linthresh' + suffix: self.symloglin,
                         'linscale' + suffix: self.symloglinscale}
        else:
            kargs = {}

        set_scale(self.scale, **kargs)

        r = (_to_float(self.range[0]), _to_float(self.range[1]))
        set_lim(r)

    def make_save_data(self, ax):
        names = ("name", "base", "auto", "range", "scale", "cmap",
                 "mode", "ticks", "symloglin", "symloglinscale")
        d = {name: getattr(self, name) for name in names}
        m = []
        for plot in self._member:
            #           m.append(ax.get_relative_path(plot()))
            m.append(ax.get_td_path(plot())[1:])
        d["member"] = m
        d["_ax_idx"] = self._ax_idx
        return d

    def import_data(self, ax, data):
        names = ("name", "base", "auto", "range", "scale", "cmap",
                 "mode", "_ax_idx", "ticks", "symloglin",
                 "symloglinscale")
        for name in names:
            if name in data:
                setattr(self, name, data[name])
        self._loaded_member = data["member"]


class AxisVisualParam(object):
    def __init__(self):
        self.minortickmode = 'off'
        self.grid = 'off'
        self.lcolor = 'black'  # label color
        self.lsize = 'default'     # label size
        self.tcolor = 'black'      # tick color
        self.otcolor = 'black'     # offset text color
        self.otsize = 'default'    # offset text size
        # text, color, font-family, weight, style, size
        self.labelinfo = ['', 'black', 'default',
                          'default', 'default', 'default']
        self.tick_position = 'left'
        self.tick_both = True
        self.box = True
        self.format = 'default'

    def set_tickparam(self, a, adir, figaxes):
        figpage = figaxes.get_figpage()
        if self.lsize == 'default':
            if figpage is not None:
                size = figpage.getp("ticklabel_size")
            else:
                size = 12
        else:
            size = self.lsize
        if self.otsize == 'default':
            if figpage is not None:
                otsize = figpage.getp("ticklabel_size")
            else:
                otsize = 12
        else:
            otsize = self.otsize
        a.tick_params(axis=adir, labelsize=size)
        a.tick_params(axis=adir, labelcolor=self.lcolor)
        a.tick_params(axis=adir, color=self.tcolor)
        # print self.otcolor
        if adir == 'x':
            a.xaxis.offsetText.set_color(self.otcolor)
            a.xaxis.offsetText.set_size(otsize)
        elif adir == 'y':
            a.yaxis.offsetText.set_color(self.otcolor)
            a.yaxis.offsetText.set_size(otsize)
        self.set_box(a)

    def set_box(self, a):
        name_pairs = {'right': 'left',
                      'top': 'bottom',
                      'left': 'right',
                      'bottom': 'top'}
        if isinstance(self.tick_position, dict):
            pass
            # custom spine setting does not use this
        else:
            if self.tick_position in name_pairs:
                if self.box:
                    name = name_pairs[self.tick_position]
                    a.spines[name].set_color('k')
                else:
                    name = name_pairs[self.tick_position]
                    a.spines[name].set_color('none')

    def _set_label(self, txt):
        txt.set_text(self.labelinfo[0])
        txt.set_color(self.labelinfo[1])
        if self.labelinfo[2] != 'default':
            txt.set_family(self.labelinfo[2])
        if self.labelinfo[3] != 'default':
            txt.set_weight(self.labelinfo[3])
        if self.labelinfo[4] != 'default':
            txt.set_style(self.labelinfo[4])
        if self.labelinfo[5] == 'default':
            # put a tentative number
            # this size will be overwritten when
            # before AxesMod::draw is called from
            # FigureMod
            txt.set_size(10)
        else:
            txt.set_size(float(self.labelinfo[5]))

    def _set_ticker(self, a):
        try:
            if not isiterable(self.ticks):
                if self.scale == 'linear':
                    a.set_major_locator(mticker.AutoLocator())
                elif self.scale == 'log':
                    a.set_major_locator(mticker.LogLocator(self.base))
                elif self.scale == 'symlog':
                    from matplotlib.scale import SymmetricalLogScale
                    if isMPL33:
                        scale = SymmetricalLogScale(a,
                                                    base=self.base,
                                                    linthresh=self.symloglin,
                                                    linscale=self.symloglinscale)
                    else:
                        scale = SymmetricalLogScale(a,
                                                    basex=self.base,
                                                    linthreshx=self.symloglin,
                                                    linscalex=self.symloglinscale)
                    a.set_major_locator(
                        mticker.SymmetricalLogLocator(scale.get_transform()))
#                    scale.set_default_locators_and_formatters(a)

                else:
                    a.set_major_locator(mticker.AutoLocator())
                #a.get_axes().locator_params(self.name[0], nbins = 10)
                if self.ticks is not None:
                    value = self.ticks
                else:
                    #figpage = a.get_axes().figobj.get_figpage()
                    figpage = a.axes.figobj.get_figpage()
                    if self.name[0] == 'x':
                        value = figpage.getp('nticks')[0]
                    elif self.name[0] == 'y':
                        value = figpage.getp('nticks')[1]
                    elif self.name[0] == 'z':
                        value = figpage.getp('nticks')[2]
                    else:
                        pass
                try:
                    # this works onlyfor MaxNLocator
                    #a.get_axes().locator_params(self.name[0], nbins = value)
                    a.axes.locator_params(self.name[0], nbins=value)
                except BaseException:
                    # for Symlog and LogLocator
                    a.get_major_locator().numticks = value
            else:
                a.set_ticks(self.ticks)

            if self.format == 'default':
                if self.scale == 'linear':
                    a.set_major_formatter(mticker.ScalarFormatter())
                elif self.scale == 'log':
                    a.set_major_formatter(
                        mticker.LogFormatterMathtext(self.base))
                elif self.scale == 'symlog':
                    a.set_major_formatter(
                        mticker.LogFormatterMathtext(self.base))
                else:
                    a.set_major_formatter(mticker.ScalarFormatter())
            elif self.format == 'scalar':
                a.set_major_formatter(mticker.ScalarFormatter())
            elif self.format == 'scalar(mathtext)':
                a.set_major_formatter(mticker.ScalarFormatter(
                    useOffset=True, useMathText=True))
                a.get_major_formatter().get_offset()
            elif self.format == 'log':
                a.set_major_formatter(mticker.LogFormatter(self.base))
            elif self.format == 'log(mathtext)':
                a.set_major_formatter(mticker.LogFormatterMathtext(self.base))
            elif self.format == 'log(exp)':
                a.set_major_formatter(mticker.LogFormatterExponent(self.base))
            elif self.format == 'none':
                a.set_major_formatter(mticker.NullFormatter())
            else:
                a.set_major_formatter(mticker.FormatStrFormatter(self.format))
        except BaseException:
            import traceback
            traceback.print_exc()

    def make_save_data(self, ax):
        names = ("minortickmode", "grid", "lcolor", "lsize", "tcolor",
                 "labelinfo", "tick_position", "tick_both", "otcolor",
                 "format", "ticks", "box", "otsize")
        return {name: getattr(self, name) for name in names}

    def import_data(self, ax, data):
        names = ("minortickmode", "grid", "lcolor", "lsize", "tcolor",
                 "labelinfo", "tick_position", "tick_both", "otcolor",
                 "format", "ticks", "box", "otsize")
        for name in names:
            if name in data:
                setattr(self, name, data[name])

        # for backward compatibility
        if not "otcolor" in data:
            setattr(self, "otcolor", data["lcolor"])
        # for backward compatibility
        if not "otsize" in data:
            setattr(self, "otsize", data["lsize"])

    def _apply_custom_spine_setting(self, spine, value):
        spine.set_position((value['loc. ref'],
                            float(value['loc. value'])))
        spine.set_alpha(value['alpha'])
        spine.set_facecolor(value['face'])
        spine.set_edgecolor(value['edge'])
        spine.set_linewidth(value['width'])
        spine.set_linestyle(value['style'])

    def _get_custom_spine_setting(self, spine):
        value = {}
        if spine.get_position() == 'zero':
            value['loc. ref'] = 'data'
            value['loc. value'] = '0'
        elif spine.get_position() == 'center':
            value['loc. ref'] = 'axes'
            value['loc. value'] = '0.5'
        else:
            value['loc. ref'] = spine.get_position()[0]
            value['loc. value'] = str(spine.get_position()[1])
        value['alpha'] = spine.get_alpha()
        value['face'] = spine.get_facecolor()
        value['edge'] = spine.get_edgecolor()
        value['width'] = spine.get_linewidth()
        value['style'] = spine.get_linestyle()
        return value


class AxisParam(AxisRangeParam, AxisVisualParam):
    def __init__(self, name):
        AxisRangeParam.__init__(self)
        AxisVisualParam.__init__(self)
        self.name = name

    def make_save_data(self, ax):
        return {"AxisRangeParam": AxisRangeParam.make_save_data(self, ax),
                "AxisVisualParam": AxisVisualParam.make_save_data(self, ax)}

    def import_data(self, ax, data):
        AxisRangeParam.import_data(self, ax, data["AxisRangeParam"])
        AxisVisualParam.import_data(self, ax, data["AxisVisualParam"])
        return self


class AxisXParam(AxisParam):
    def __init__(self, *args, **kargs):
        AxisParam.__init__(self, *args, **kargs)
        self.tick_position = 'bottom'

    def set_tickparam(self, a, fa):
        AxisVisualParam.set_tickparam(self, a, 'x', fa)
        self.set_artist_tickposition(a)
        if self.tick_both:
            a.xaxis.set_ticks_position('both')

    def get_artist_tickposition(self, a):
        if isinstance(self.tick_position, dict):
            return self.tick_position
        else:
            value = {}
            def_set = ['top', 'bottom']
            for x, name in zip(def_set, ['spine1', 'spine2']):
                value[name] = self._get_custom_spine_setting(a.spines[x])

            value['label'] = self.tick_position
            return value

    def set_artist_tickposition(self, a):
        value = self.tick_position
        def_set = ['top', 'bottom']
        if isinstance(self.tick_position, dict):
            a.xaxis.set_ticks_position(value['label'])
            for x, name in zip(def_set, ['spine1', 'spine2']):
                self._apply_custom_spine_setting(a.spines[x], value[name])
        else:
            if value in def_set:
                a.xaxis.set_ticks_position(value)
                for x in def_set:
                    a.spines[x].set_position(('outward', 0))
            elif value == 'zero':
                a.xaxis.set_ticks_position(def_set[0])
                for x in def_set:
                    a.spines[x].set_position('zero')
            elif value == 'center':
                a.xaxis.set_ticks_position(def_set[0])
                for x in def_set:
                    a.spines[x].set_position('center')
            else:
                pass

    def set_label(self, a):
        txt = a.get_xaxis().label
        self._set_label(txt)

    def get_label_from_mpl(self, a):
        return a.get_xaxis().label.get_text()

    def set_ticks(self, a):
        xaxis = a.get_xaxis()
        self._set_ticker(xaxis)

    def get_mpl_range(self, a):
        return a.get_xlim()

    def set_artist_rangeparam(self, a):

        set_scale = a.set_xscale
        set_auto = a.set_autoscalex_on
        set_lim = a.set_xlim
        AxisRangeParam.set_artist_rangeparam(self,
                                             set_scale=set_scale,
                                             set_auto=set_auto,
                                             set_lim=set_lim)
        self.set_ticks(a)


class AxisYParam(AxisParam):
    def __init__(self, *args, **kargs):
        AxisParam.__init__(self, *args, **kargs)
        self.tick_position = 'left'

    def set_tickparam(self, a, fa):
        AxisVisualParam.set_tickparam(self, a, 'y', fa)
        self.set_artist_tickposition(a)
        if self.tick_both:
            a.yaxis.set_ticks_position('both')

    def get_artist_tickposition(self, a):
        if isinstance(self.tick_position, dict):
            return self.tick_position
        else:
            value = {}
            def_set = ['left', 'right']
            for x, name in zip(def_set, ['spine1', 'spine2']):
                value[name] = self._get_custom_spine_setting(a.spines[x])

            value['label'] = self.tick_position
            return value

    def set_artist_tickposition(self, a):
        value = self.tick_position
        def_set = ['left', 'right']
        if isinstance(self.tick_position, dict):
            a.yaxis.set_ticks_position(value['label'])
            for x, name in zip(def_set, ['spine1', 'spine2']):
                self._apply_custom_spine_setting(a.spines[x], value[name])
        else:
            if value in def_set:
                a.yaxis.set_ticks_position(value)
                for x in def_set:
                    a.spines[x].set_position(('outward', 0))
            elif value == 'zero':
                a.yaxis.set_ticks_position(def_set[0])
                for x in def_set:
                    a.spines[x].set_position('zero')
            elif value == 'center':
                a.yaxis.set_ticks_position(def_set[0])
                for x in def_set:
                    a.spines[x].set_position('center')
            else:
                pass

    def set_label(self, a):
        txt = a.get_yaxis().label
        self._set_label(txt)

    def get_label_from_mpl(self, a):
        return a.get_yaxis().label.get_text()

    def set_ticks(self, a):
        yaxis = a.get_yaxis()
        self._set_ticker(yaxis)

    def get_mpl_range(self, a):
        return a.get_ylim()

    def set_artist_rangeparam(self, a):
        set_scale = a.set_yscale
        set_auto = a.set_autoscaley_on
        set_lim = a.set_ylim
        AxisRangeParam.set_artist_rangeparam(self,
                                             set_scale=set_scale,
                                             set_auto=set_auto,
                                             set_lim=set_lim)
        self.set_ticks(a)


class AxisZParam(AxisParam):
    def __init__(self, *args, **kargs):
        AxisParam.__init__(self, *args, **kargs)
        self.tick_position = 'default'  # not used??

    def set_artist_axisparam(self, a):
        pass

    def set_tickparam(self, a, fa):
        AxisVisualParam.set_tickparam(self, a, 'z', fa)

    def set_label(self, a):
        txt = a.get_zaxis().label
        self._set_label(txt)

    def get_label_from_mpl(self, a):
        return a.get_zaxis().label.get_text()

    def set_ticks(self, a):
        zaxis = a.get_zaxis()
        self._set_ticker(zaxis)

    def get_mpl_range(self, a):
        return a.get_zlim()

    def set_artist_rangeparam(self, a):
        if not hasattr(a, 'set_zlim'):
            return

        set_scale = a.set_zscale

        def func(value, ax=a):
            ax.autoscale(value, axis='z')
        set_auto = func
        set_auto = a.set_autoscalez_on
        set_lim = a.set_zlim
        AxisRangeParam.set_artist_rangeparam(self,
                                             set_scale=set_scale,
                                             set_auto=set_auto,
                                             set_lim=set_lim)
        self.set_ticks(a)


class AxisCParam(AxisParam):
    def __init__(self, name):
        AxisParam.__init__(self, name)
        self.tick_position = 'default'  # not used??
        self._cb = None
        self._cm = None
        self.clip = ((False, 'white'), (False, 'white'))

    def set_tickparam(self, a, fa):
        pass

    def set_label(self, a):
        pass

    def get_label_from_mpl(self, a):
        return self.labelinfo[0]

    def set_ticks(self, a):
        pass

    def get_mpl_range(self, a):
        if isinstalce(a, ScalarMappable):
            return a.get_clim()
        else:
            return (0, 1)

    def set_artist_rangeparam(self, a):
        '''
        a is axes object
        need to find all Scalarmappable
        '''

        figobjs = [a2.figobj for a2 in a.get_children()
                   if hasattr(a2, 'figobj')]
        for figobj in set(figobjs):
            for a2 in figobj.get_mappable():
                self.set_crangeparam_to_artist(a2)

    def set_crangeparam_to_artist(self, a2, check=True):
        '''
        a2 is artist
        '''
        if check:
            for m in self._member:
                if a2 in m().get_mappable():
                    break
            else:
                return

        if self._cm is None:
            self._cm = get_cmap(self.cmap, 256)
        cm = self._cm

        if self.clip[0][0]:
            cm.set_under(self.clip[0][1])
        else:
            cm.set_under(cm(0))
        if self.clip[1][0]:
            cm.set_over(self.clip[1][1])
        else:
            cm.set_over(cm(cm.N - 1))
        a2.set_cmap(cm)

        if self.scale == 'linear':
            a2.set_norm(Normalize(self.range[0],
                                  self.range[1]))
            a2.set_clim(self.range)
        elif self.scale == 'symlog':
            a2.set_norm(SymLogNorm(self.symloglin,
                                   vmin=self.range[0],
                                   vmax=self.range[1],
                                   base=10)
                        )
            a2.set_clim(self.range)
        else:
            a2.set_norm(LogNorm(self.range[0],
                                self.range[1]))
            a2.set_clim(self.range)

    def show_cbar(self, figaxes,
                  offset=0.0,
                  position=(0.9, 0.1),
                  size=(0.05, 0.8),
                  lsize='default',
                        lcolor='black',
                  olsize='default',
                        olcolor='black',
                  direction='v'):

        if self._cb is None:
            ichild = figaxes.add_colorbar()
            cb = figaxes.get_child(ichild)
            anchor = cb.getp("inset_anchor")
            anchor = (position[0] + offset, position[1])
            cb.setp("inset_anchor", anchor)
            cb.setp("cdir", direction)
            cb.setp("inset_w", size[0])
            cb.setp("inset_h", size[1])
            cb.set_caxis_param(self)
            cb.realize()
            self._cb = weakref.ref(cb, self._cb_dead)

            if direction == 'v':
                name = 'y'
            else:
                name = 'x'

            ap = cb.get_axis_param(name)
            artist = cb.get_axes_artist_by_name(name)[0]
            ap.lsize = lsize
            ap.otsize = olsize
            ap.lcolor = lcolor
            ap.otcolor = olcolor
            ap.set_tickparam(artist, cb)
            # set cb params here...

    def hide_cbar(self):
        if self._cb is None:
            return
        if self._cb() is None:
            return
        cb = self._cb()
        figaxes = cb.get_parent()
        cb.destroy()
        self._cb = None
        ifigure.events.SendChangedEvent(figaxes, w=None)

    def has_cbar(self):
        if self._cb is None:
            return False
        if self._cb() is None:
            return False
        return True

    def cbar_shown(self):
        if self._cb is not None:
            if self._cb() is not None:
                return True
            else:
                self._cb = None
                return False
        return False

    def _cb_dead(self, v):
        self._cb = None

    def rm_member(self, obj):
        super(AxisCParam, self).rm_member(obj)
        if len(self._member) == 0:
            if self._cb is None:
                return
            if self._cb() is not None:
                figaxes = self._cb().get_parent()
                self._cb().destroy()
                ifigure.events.SendChangedEvent(figaxes, w=None)
            self._cb = None

    def set_cmap(self, cmap):
        self.cmap = cmap
        self._cm = get_cmap(self.cmap, 256)

    def get_cmap(self):
        return self.cmap

    def update_cb(self):
        if self._cb is None:
            return
        if self._cb() is None:
            return
        self._cb().update_cbar_image()
        self._cb().set_bmp_update(False)

    def make_save_data(self, ax):
        d = AxisParam.make_save_data(self, ax)
        d["clip"] = self.clip
        return d

    def import_data(self, ax, data):
        AxisParam.import_data(self, ax, data)
        self.clip = data["clip"]
        return self

#    def add_colorbar(self, name=None, *args, **kywds):



from ifigure.widgets.canvas.ifigure_canvas import ifigure_canvas
from .fig_mds import FigMds
from ifigure.widgets.undo_redo_history import GlobalHistory
from ifigure.widgets.undo_redo_history import UndoRedoArtistProperty
from ifigure.widgets.undo_redo_history import UndoRedoFigobjProperty
from ifigure.widgets.undo_redo_history import UndoRedoFigobjMethod
from ifigure.widgets.undo_redo_history import UndoRedoGroupUngroupFigobj
from ifigure.widgets.undo_redo_history import UndoRedoAddRemoveArtists


def get_figmds(obj):
    p = obj.get_parent()
    while p is not None:
        if isinstance(p, FigMds):
            return p
        p = p.get_parent()
    return p


class MDSScopeCanvas(ifigure_canvas):
    def change_figobj_axes(self, figobj, value, direction):
        if len(self.selection) == 0:
            return
        h = []
        if direction != 'c':
            for a in self.selection:
                h.append(UndoRedoFigobjMethod(a(), 'container_idx', value))
            figmds = get_figmds(a().figobj)
            if figmds is not None:
                if len(figmds._artists) > 0:
                    h.append(UndoRedoFigobjMethod(
                        figmds._artists[0], 'container_idx', value))
        else:
            for a in self.selection:
                h.append(UndoRedoFigobjMethod(a(), 'caxis_idx', value))
        ax = a().axes
        h.append(UndoRedoFigobjMethod(ax, 'adjustrange', None))

        window = self.GetTopLevelParent()
        GlobalHistory().get_history(window).make_entry(h, use_reverse=False)



from ifigure.widgets.dlg_preference import PrefComponent
from ifigure.utils.setting_parser import iFigureSettingParser as SP
import wx
import ifigure.utils.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('HelperApp')


def config_name():
    if wx.Platform == '__WXMAC__':
        return 'pref.helperapp_mac_config'
    else:
        return 'pref.helperapp_config'


class HelperApp(PrefComponent):
    def __init__(self):
        PrefComponent.__init__(self, 'HelperApp')
        p = SP()

        self.setting = p.read_setting(config_name())

    def save_setting(self):
        p = SP()
        p.write_setting(config_name(), self.setting)

    def get_dialoglist(self):
        txt = 'yes' if self.setting["use_editor"] else 'no'

        list1 = [["Pdf",         str(self.setting["pdf"]), 200, None],
                 ["PostScript",  str(self.setting["ps"]), 200, None],
                 ["Editor",      str(self.setting["editor"]), 200, None],
                 ["ghostscript", str(self.setting["gs"]), 200, None],
                 ["ps2pdf",      str(self.setting["ps2pdf"]), 200, None],
                 ["convert",     str(self.setting["convert"]), 200, None],
                 ["Use external editor",  txt, 1, {"values": ["yes", "no"]}],
                 [None, "Replacement rule: {bin} -> piscope_sourcer/bin, {0} -> file path"+" "*30, 2], ]
        hint1 = ['PDF viewer', 'Postscript viewer', 'Externail Editor',
                 'gs', 'ps2pdf(used in fig_eps)', None, None]
        return list1, hint1

    def set_dialog_result(self, value):
        self.setting["pdf"] = str(value[0])
        self.setting["ps"] = str(value[1])
        self.setting["editor"] = str(value[2])
        self.setting["gs"] = str(value[3])
        self.setting["ps2pdf"] = str(value[4])
        self.setting["convert"] = str(value[5])
        self.setting["use_editor"] = (str(value[6]) == 'yes')



from __future__ import print_function
#
#   run script for StandardSolver
#
#   make a thread to wait for the result,
#   and the thread submit a job, which is started
#   in a main thread.
#
from six.moves import queue as Queue
import threading
import weakref
import wx
renew_model_each_time = True
solver = obj.get_parent()
tmodel = solver._model
use_def_merger = solver._use_def_merger


def run_workers(workers, sol):
    def submit(worker, queue):
        rpath = tmodel.get_root_model().get_td_path(tmodel)
        rname = tmodel.get_root_model()._name
        rmodel = worker.get_child(name=rname)

        if (renew_model_each_time and
                rmodel is not None):
            # print "Desotry", rmodel
            rmodel.destroy()
        rmodel = worker.get_child(name=rname)
        if rmodel is None:
            tmodel.get_root_model().duplicate(worker, rname, save_script_link=True)
            rmodel = worker.get_child(name=rname)

        model = rmodel.resolve_td_path(rpath)
        threads = model.Run(return_queue=queue)
        # print 'running ', model.get_full_path(), threads[0]
        worker._thread_name = threads[0]

    def finish_job(method, idx, w, sol, queue):
        method(idx, w, sol)
        queue.put(w)
        solver._queue = None

    q = Queue.Queue()
    solver._queue = q
    wx.CallAfter(submit, *(w, q))
    while True:
        try:
            t = q.get(True)
            break
        except Queue.Empty:
            solver._queue = None
            print('Failed to get response')
            return
        # print 'response from', w, threading.current_thread()
        if t == 'abort job':
            # for a moment let's ignore this...
            continue

    rpath = tmodel.get_root_model().get_td_path(tmodel)
    rname = tmodel.get_root_model()._name
    rmodel = w.get_child(name=rname)
    model = rmodel.resolve_td_path(rpath)

    if (model._finish_script == ')' or
            use_def_merger):
        model._finish_script = solver.merge_sol.get_full_path()
    # print 'calling merge_sol', model.do_finish
    wx.CallAfter(finish_job, model.do_finish, 0, w, sol, q)
#    model.do_finish(0, w, sol)


sol_base = solver._sol
workers = solver.get_workers()
if len(workers) == 0:
    print('no workers')
    stop()
if solver._queue is not None:
    print('there is a waiting queue, exiting...')
    stop()
for w in workers:
    w._thread_name = ''
t = threading.Thread(target=run_workers, args=(workers, sol_base))
wx.CallAfter(t.start)



import wx
from ifigure.utils.setting_parser import iFigureSettingParser as SP
from ifigure.utils.edit_list import EditListPanel
import ifigure.widgets.dialog as dialog

pref = 'pref.file_helper'


class FileHelperDialog(wx.Dialog):
    def __init__(self, parent, id=-1, title="File Action Setting"):

        p = SP().set_rule('file',
                          {'name': '', 'ext': '', 'action': '',
                           'action_txt': '', 'use': False})
        self.var = p.read_setting(pref)

        wx.Dialog.__init__(self, parent, id, title,
                           style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)

        self.mainSizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.mainSizer)

        self.cb = wx.ComboBox(self, wx.ID_ANY,
                              size=(-1, -1),
                              style=wx.CB_READONLY,
                              choices=[x['name'] for x in self.var['file']])
        self.mainSizer.Add(self.cb, 0, wx.ALL, 3)

        l = [["rule",   '', 200, None],
             ["action", '', 235, {'nlines': 3}],
             ["note", '', 0, None],
             [None, False, 3,  {'text': 'Use'}], ]

        self.elp = EditListPanel(self, l, edge=0, call_sendevent=self)
        self.elp.GetSizer().AddGrowableRow(1)

        self.mainSizer.Add(self.elp, 1, wx.ALL | wx.EXPAND, 2)

        self.buttonSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.bnew = wx.Button(self, label="New...", id=wx.ID_ANY)
        self.bdel = wx.Button(self, label="Delete...", id=wx.ID_ANY)
        self.bok = wx.Button(self, label="OK", id=wx.ID_OK)

        self.buttonSizer.Add(self.bnew, 0, wx.ALL, 7)
        self.buttonSizer.Add(self.bdel, 0, wx.ALL, 7)
        self.buttonSizer.AddStretchSpacer(1)
        self.buttonSizer.Add(self.bok, 0, wx.ALL, 7)
        self.mainSizer.Add(self.buttonSizer, 0, wx.ALL | wx.EXPAND, 0)
        self.Bind(wx.EVT_BUTTON, self.onOk, id=wx.ID_OK)
        self.Bind(wx.EVT_BUTTON, self.onNew, self.bnew)
        self.Bind(wx.EVT_BUTTON, self.onDel, self.bdel)
        self.cb.Bind(wx.EVT_COMBOBOX, self.onCBHit)
        wx.CallAfter(self.Layout)

        self.set_panels_2_index0()

    def send_event(self, elp, evt0):
        for v in self.var['file']:
            if v['name'] == str(self.cb.GetValue()):
                value = self.elp.GetValue()
                v['ext'] = value[0]
                v['action'] = value[1]
                v['action_txt'] = value[2]
                v['use'] = value[3]
                break

    def set_elp(self):
        vv = self.var['file']
        for v in vv:
            if v['name'] == str(self.cb.GetValue()):
                self.elp.SetValue((str(v['ext']),
                                   str(v['action']),
                                   str(v['action_txt']), v['use']))
                break

    def update_cb(self):
        names = self.get_names()
        self.cb.Clear()
        for x in names:
            self.cb.Append(x)

    def get_names(self):
        names = [x['name'] for x in self.var['file']]
        return names

    def onCBHit(self, evt):
        for v in self.var['file']:
            if v['name'] == self._cb_value:
                value = self.elp.GetValue()
                v['ext'] = value[0]
                v['action'] = value[1]
                v['action_txt'] = value[2]
                v['use'] = value[3]
        self.set_elp()
        self._cb_value = str(self.cb.GetValue())

    def onOk(self, evt):
        for v in self.var['file']:
            if v['name'] == self._cb_value:
                value = self.elp.GetValue()
                v['ext'] = value[0]
                v['action'] = value[1]
                v['action_txt'] = value[2]
                v['use'] = value[3]
        if self.IsModal():
            self.EndModal(wx.ID_OK)
        else:
            self.Destroy()

    def onDel(self, evt):
        vv = self.var['file']
        self.var['file'] = [v for v in vv if v['name']
                            != str(self.cb.GetValue())]
        self.set_panels_2_index0()

    def set_panels_2_index0(self):
        names = self.get_names()
        self.update_cb()
        self.cb.SetValue(names[0])
        self._cb_value = names[0]
        self.set_elp()

    def onNew(self, evt):
        ret, new_name = dialog.textentry(self,
                                         "Enter the name of file type",
                                         "Add New FileType", '')
        if not ret:
            return
        if not new_name in self.get_names():
            v = {'name': new_name, 'ext': '*.*',
                 'action': '', 'action_txt': '', 'use': True}
        self.var['file'].append(v)
        self.update_cb()
        self.cb.SetValue(new_name)
        self._cb_value = new_name
        self.set_elp()


class Handler(object):
    def __init__(self, parent):
        pass

    def on_button(self, ev):
        parent_window = ev.GetEventObject().GetTopLevelParent()
        dlg = FileHelperDialog(parent_window)
        a = dlg.ShowModal()
        if a == wx.ID_OK:
            p = SP().set_rule('file',
                              {'name': '', 'ext': '', 'action': '',
                               'action_txt': '', 'use': False})
            p.write_setting(pref, dlg.var)

    def get_value(self):
        return None

    def set_value(self, v):
        pass


def get_handler(parent):
    return Handler(parent)



from __future__ import print_function


'''
PostOffice 
simple postoffice to send a file
'''

import wx
import binascii
#import sys, socket, shlex, threading, os, binascii, subprocess, time, tempfile
from ifigure.widgets.passwd_dialog import PasswordDialog, UsernamePasswordDialog
from ifigure.utils.setting_parser import iFigureSettingParser as SP
from ifigure.utils.edit_list import DialogEditList
from ifigure.utils.mailfile import checkPasswd, sendMail
from ifigure.widgets.dlg_preference import PrefComponent
import ifigure.utils.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('PostOffice')

coding = 'utf-8'

class PostOffice(PrefComponent):
    def __init__(self):
        PrefComponent.__init__(self, 'PostOffice')
        p = SP()
        self.setting = p.read_setting('pref.postoffice_config')
        self.passwd = None
        self.username = None
        self.parent = None

    def save_setting(self):
        p = SP()
        p.write_setting('pref.postoffice_config', self.setting)

    def set_parent(self, parent):
        self.parent = parent
        return self

    def ask_passwd(self, parent=None):
        if parent is None:
            parent = self.parent
        p_dlg = UsernamePasswordDialog(parent,
                                       title="Mail Server Connection (SSL)",
                                       label=("Enter Name/Password for SMPT server (" +
                                              self.setting["server"] + ':' + str(self.setting["ssl_port"])+")"))
        a = p_dlg.ShowModal()
        if a == wx.ID_OK:
            if checkPasswd(self.setting["server"],
                           ssl_port=self.setting["ssl_port"],
                           ssl_username=p_dlg.result[0],
                           ssl_passwd=p_dlg.result[1]):
                self.username = p_dlg.result[0]
                self.passwd = binascii.b2a_hex(p_dlg.result[1].encode(coding))
            else:
                print('Wrong username/password')
        p_dlg.Destroy()

    def get_dialoglist(self):
        lists = self._elp_panel_list()
        hints = [None]*len(lists)
        return lists, hints

    def set_dialog_result(self, data):
        self._elp_process_panelvalue(data)

    def edit_setting(self):
        list6 = self._elp_panel_list()
        value = DialogEditList(list6, modal=True,
                               style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER,
                               tip=None,
                               parent=None,)
        if value[0] is True:
            self._elp_process_panelvalue(value[1])

    def _elp_panel_list(self):
        list6 = [["Subject (default)", self.setting["default_title"], 0, None],
                 ["Message (default)", self.setting["default_message"], 0, None],
                 ["To (default)", self.setting["default_to"], 0, None],
                 ["Mail server", self.setting["server"], 0, None],
                 ["Use ssl", self.setting["use_ssl"],
                     1, {"values": ["on", "off"]}],
                 ["Ssl port", str(self.setting["ssl_port"]), 0, None], ]
        return list6

    def _elp_process_panelvalue(self, value):
        self.setting["default_title"] = str(value[0])
        self.setting["default_message"] = str(value[1])
        self.setting["default_to"] = str(value[2])
        self.setting["server"] = str(value[3])
        self.setting["use_ssl"] = str(value[4])
        self.setting["ssl_port"] = int(value[5])

    def send_message(self, message='', parent=None):
        if parent is None:
            parent = self.parent
        if self.passwd is None:
            self.ask_passwd(parent=parent)
        if self.passwd is None:
            return

        list6 = [["subject", self.setting["default_title"], 0, None],
                 ["to", self.setting["default_to"], 0, None], ]
#                 ["mail server", self.setting["server"], 0, None],
#                 ["use ssl", self.setting["use_ssl"], 1, {"values":["on", "off"]}],
#                 ["ssl port", str(self.setting["ssl_port"]), 0, None],]

        value = DialogEditList(list6, modal=True,
                               style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER,
                               tip=None,
                               parent=parent,)

        if value[0] is True:
            subject = str(value[1][0])
            message = message
            to = str(value[1][1])
        else:
            return

        passwd = binascii.a2b_hex(self.passwd)
        sendMail([to], subject, message, files=[],
                 server=self.setting["server"],
                 ssl=(self.setting["use_ssl"] == "on"),
                 ssl_port=int(self.setting["ssl_port"]),
                 ssl_username=self.username,
                 ssl_passwd=passwd)

    def send_file(self, file, parent=None):
        if parent is None:
            parent = self.parent
        if self.passwd is None:
            self.ask_passwd(parent=parent)
        if self.passwd is None:
            return

        list6 = [["To:", self.setting["default_to"], 0, None],
                 ["Subject:", self.setting["default_title"], 0, None],
                 [None, self.setting["default_message"], 35, {'nlines': 10}], ]

#                 ["mail server", self.setting["server"], 0, None],
#                 ["use ssl", self.setting["use_ssl"], 1, {"values":["on", "off"]}],
#                 ["ssl port", str(self.setting["ssl_port"]), 0, None],]

        value = DialogEditList(list6, modal=True,
                               style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER,
                               tip=None, title='Edit Message for Send File',
                               parent=parent, size=(400, -1))

        if value[0] is True:
            to = str(value[1][0])
            subject = str(value[1][1])
            message = str(value[1][2])

        else:
            return

        passwd = binascii.a2b_hex(self.passwd).decode(coding)
        
        sendMail([to], subject, message, files=[file],
                 server=self.setting["server"],
                 ssl=(self.setting["use_ssl"] == "on"),
                 ssl_port=int(self.setting["ssl_port"]),
                 ssl_username=self.username,
                 ssl_passwd=passwd)



from __future__ import print_function
#
#
#  this is a start-up script loaded in
#  shell window.
#  This script defines functions available in shell and
#  help function
#
#
import six
import numpy as np
from ifigure.interactive import aviewer
from ifigure.interactive import plot, timetrace, oplot, plotc, loglog, semilogx, semilogy
from ifigure.interactive import errorbar, errorbarc, oerrorbar
from ifigure.interactive import triplot, delaunay
from ifigure.interactive import fill, fill_between, fill_betweenx, fill_between_3d
from ifigure.interactive import contour, contourf
from ifigure.interactive import image, spec, specgram
from ifigure.interactive import scatter, hist
from ifigure.interactive import tripcolor, tricontour, tricontourf
from ifigure.interactive import axline, axlinec
from ifigure.interactive import axspan, axspanc
from ifigure.interactive import surf, solid, trisurf, revolve
from ifigure.interactive import quiver, quiver3d
from ifigure.interactive import annotate
from ifigure.interactive import legend
from ifigure.interactive import text, figtext, arrow, figarrow
from ifigure.interactive import ispline
from ifigure.interactive import figure, close, edit
from ifigure.interactive import newbook
from ifigure.interactive import scope, scopenw, tscope
from ifigure.interactive import video, waveviewer, videoviewer
#from ifigure.interactive import sliceviewer
from ifigure.interactive import addpage
from ifigure.interactive import delpage
from ifigure.interactive import hold
from ifigure.interactive import cla
from ifigure.interactive import cls
from ifigure.interactive import clf
from ifigure.interactive import nsection
from ifigure.interactive import nsec
from ifigure.interactive import subplot
from ifigure.interactive import isection
from ifigure.interactive import isec
from ifigure.interactive import showpage
from ifigure.interactive import title, suptitle
from ifigure.interactive import xlabel, ylabel, zlabel, clabel
from ifigure.interactive import xtitle, ytitle, ztitle, ctitle
from ifigure.interactive import update
from ifigure.interactive import draw
from ifigure.interactive import server
from ifigure.interactive import clear
from ifigure.interactive import debug
from ifigure.interactive import profile, profile_start, profile_stop
from ifigure.interactive import threed, lighting
#from ifigure.interactive import newcz
from ifigure.interactive import ipage
from ifigure.interactive import twinx
from ifigure.interactive import twiny
from ifigure.interactive import twinc
from ifigure.interactive import exportv
from ifigure.interactive import importv
from ifigure.interactive import xlog, ylog, zlog, clog
from ifigure.interactive import xsymlog, ysymlog, zsymlog, csymlog
from ifigure.interactive import xlinear, ylinear, zlinear, clinear
from ifigure.interactive import xlim, ylim, zlim, clim
from ifigure.interactive import xauto, yauto, zauto, cauto
from ifigure.interactive import xnames, ynames, znames, cnames
from ifigure.interactive import autoplay, cbar
from ifigure.interactive import autoplay, setupmodel
from ifigure.interactive import property
from ifigure.interactive import view
from ifigure.interactive import glinfo, savefig, savedata

if six.PY3:
    from importlib import reload
    from ifigure.interactive import futurize
else:
    from builtins import reload

from ifigure.interactive import has_petra
if has_petra:
    from ifigure.interactive import petram


def help(*args):
    '''
    help : command-line access to __doc__
    help() : show all availabe functions
    help(object) : show __doc__ of object

    (example)
    help(plot) : show help for plot function    
    help(help) : show this help.
    '''

    if len(args) == 0:
        import types
        lvar = locals()
        gvar = globals()
        lnames = [key for key in lvar if type(lvar[key]) == types.FunctionType]
        gnames = [key for key in gvar if type(gvar[key]) == types.FunctionType]
        print('Availabe functions')
#       print lnames
#       print gnames
        txt = ['    ']
        gnames.append('quit')
        gnames.append('forcequit')
        gnames = sorted(gnames)
        #gnames = [name for name in gnames if not name in _internal_func]
        flen = max([len(name) for name in gnames])+3
        f = '{:'+str(flen)+'}'
        gnames = [f.format(name) for name in gnames]

        # this is not to show some functions in list
        for name in gnames:
            if name.startswith('_'):
                continue

            if len(txt[-1])+len(name) > 80:
                #              txt[-1] = txt[-1]
                txt.append('    ')
            txt[-1] = '\t'.join([txt[-1], name])
        txt.append(
            '! execute shell command (note: do not call interactive command such as python/bash from this)')
        print(('\n'.join(txt)))

    else:
        f = args[0]
        if hasattr(f, '__doc__'):
            if f.__doc__ is not None:
                print((f.__doc__))
            else:
                print('help is not available for ' + f.__repr__())
        else:
            print(('help is not available for ', f))



#!/usr/bin/env python
from __future__ import print_function
import os
import sys
import smtplib
import six
if six.PY2:
    from email.Utils import formatdate
    from email import Encoders    
else:
    from email.utils import formatdate
    from email import encoders as Encoders        
from six.moves.email_mime_multipart import MIMEMultipart
from six.moves.email_mime_base import MIMEBase
from six.moves.email_mime_text import MIMEText

COMMASPACE = ', '

def checkPasswd(server, ssl_port=465, ssl_username='', ssl_passwd=''):
    try:
        smtp = smtplib.SMTP_SSL(server, port=ssl_port)
        smtp.login(ssl_username, ssl_passwd)
        smtp.close()
    except smtplib.SMTPAuthenticationError:
        return False
    return True


def sendMail(to, subject, text, files=[], server="mail1.psfc.mit.edu",
             ssl=False, ssl_port=465, ssl_username='', ssl_passwd=''):
    assert type(to) == list
    assert type(files) == list
#    me =  'shiraiwa@psfc.mit.edu'
    me = ssl_username  # 'shiraiwa@psfc.mit.edu'
    msg = MIMEMultipart()
    msg['From'] = me
#    msg['To'] = COMMASPACE.join(['shiraiwa@psfc.mit.edu'])
    msg['To'] = COMMASPACE.join(to)
    msg['Date'] = formatdate(localtime=True)
    msg['Subject'] = subject

    msg.attach(MIMEText(text))

    for file in files:
        part = MIMEBase('application', "octet-stream")
        part.set_payload(open(file, "rb").read())
        Encoders.encode_base64(part)
        part.add_header('Content-Disposition', 'attachment; filename="%s"'
                        % os.path.basename(file))
        msg.attach(part)

    if ssl:
        print((server, ssl_port))
        smtp = smtplib.SMTP_SSL(server, ssl_port)
        smtp.login(ssl_username, ssl_passwd)
    else:
        smtp = smtplib.SMTP(server)
    smtp.sendmail(me, to, msg.as_string())
    smtp.close()


if __name__ == '__main__':
    if len(sys.argv[1:]) == 1:
        file = sys.argv[1:][0]
        if os.path.exists(file):
            print('sending : '+file)
            file = os.path.abspath(file)
        sendMail(
            ["shiraiwa@psfc.mit.edu"],
            "mail attachement", "attached!\n",
            [file]
        )



from __future__ import print_function

'''
   connections is a board game from 80's

   usage:
      from ifigure.extra.connections import new_game
      new_game()

   DOTO:
      develop an algorithm to decide next move in 
      order to implement one player mode.
'''

from ifigure.widgets.book_viewer import BookViewer
from matplotlib.patches import Rectangle
import wx
import numpy as np
from matplotlib.patches import PathPatch

# size of elements
a = 1./(2**(0.5))
b = 1.0
d = 1./(11*b + 10 * a)
params = None


class Params(object):
    def __init__(self):
        self.cur_dir = 'h'
        self.cur_color = 'black'
        self.cur_pos = [0, 0]
        self.b_pos = []
        self.r_pos = []
        self.mode_mode = 1
        self._last_move = None
        self._last_obj = None


def draw_field():
    page = params.page
    figure = page._artists[0]
    step = (2*a + 2*b)*d
    xy1 = [((a+b)*d+step*x, 0) for x in range(5)]
    xy2 = [(step*x, (a+b)*d) for x in range(6)]

    for m in range(6):
        for xy in xy1:
            hlp = Rectangle((xy[0], xy[1]+m*step), b*d, b*d, facecolor='r',
                            figure=figure, transform=figure.transFigure)
            figure.patches.append(hlp)
            hlp = Rectangle((xy[1]+m*step, xy[0]), b*d, b*d, facecolor='k',
                            figure=figure, transform=figure.transFigure)
            figure.patches.append(hlp)
    viewer = wx.GetApp().TopWindow.find_bookviewer(page.get_figbook())
    viewer.draw()


def draw_octagon(loc, color='black', direction='v', alpha=1):
    ## loc (0,0) - (8,8)
    # loc[1] - loc[0] shoud be even
    import matplotlib.path as mpath
    from numpy import cos, sin, sqrt
    Path = mpath.Path
    page = params.page

    xy = [(a + 1.5*b)*d + l*(a+b)*d for l in loc]
    r = sqrt((d*b)**2 + ((a+a+b)*d)**2)/2
    figure = page._artists[0]
    theta = np.array([(22.5 + 45.*x)*np.pi/180 for x in range(8)])

    theta2 = theta[[7, 0, 1, 2, 3, 4, 5, 6, 7]]
    verts = [(xy[0]+r*cos(t), xy[1]+r*sin(t)) for t in theta2]
    codes = [Path.MOVETO] + [Path.LINETO]*(len(verts)-1)
    path = mpath.Path(verts, codes)
    obj1 = PathPatch(path, facecolor='white', edgecolor='black', alpha=alpha,
                     figure=figure, transform=figure.transFigure)

    if direction == 'v':
        theta2 = theta[[6, 1, 2, 5, 6]]
    else:
        theta2 = theta[[7, 0, 3, 4, 7]]
    verts = [(xy[0]+r*cos(t), xy[1]+r*sin(t)) for t in theta2]
    codes = [Path.MOVETO] + [Path.LINETO]*(len(verts)-1)
    path = mpath.Path(verts, codes)
    obj2 = PathPatch(path, edgecolor='black', alpha=alpha,
                     facecolor=color, figure=figure, transform=figure.transFigure)
    figure.patches.extend([obj1, obj2])
    viewer = wx.GetApp().TopWindow.find_bookviewer(page.get_figbook())
    viewer.draw()
    return obj1, obj2


def draw_hl_octagon(loc, color='cyan', alpha=1):
    ## loc (0,0) - (8,8)
    # loc[1] - loc[0] shoud be even
    import matplotlib.path as mpath
    from numpy import cos, sin, sqrt
    Path = mpath.Path
    page = params.page

    xy = [(a + 1.5*b)*d + l*(a+b)*d for l in loc]
    r = sqrt((d*b)**2 + ((a+a+b)*d)**2)/2
    figure = page._artists[0]
    theta = np.array([(22.5 + 45.*x)*np.pi/180 for x in range(8)])

    theta2 = theta[[7, 0, 1, 2, 3, 4, 5, 6, 7]]
    verts = [(xy[0]+r*cos(t), xy[1]+r*sin(t)) for t in theta2]
    codes = [Path.MOVETO] + [Path.LINETO]*(len(verts)-1)
    path = mpath.Path(verts, codes)
    obj1 = PathPatch(path, facecolor=color, edgecolor=color, alpha=alpha,
                     figure=figure, transform=figure.transFigure,)
    figure.patches.extend([obj1])
    return obj1


class Connections(BookViewer):
    pass


def draw_cursor():
    p = globals()['params']
    for obj in p.cursors:
        p.page._artists[0].patches.remove(obj)
    obj1, obj2 = draw_octagon(p.cur_pos, color=p.cur_color,
                              direction=p.cur_dir, alpha=0.4)
    p.cursors = [obj1, obj2]


def handle_keyevent(evt):
    import wx
    p = globals()['params']
    if (evt.GetKeyCode() == wx.WXK_RIGHT and
            p.cur_pos[0] < 7):
        p.cur_pos = [p.cur_pos[0]+2, p.cur_pos[1]]
    elif (evt.GetKeyCode() == wx.WXK_LEFT and
          p.cur_pos[0] > 1):
        p.cur_pos = [p.cur_pos[0]-2, p.cur_pos[1]]
    elif (evt.GetKeyCode() == wx.WXK_UP and
          p.cur_pos[1] < 7):
        p.cur_pos = [p.cur_pos[0], p.cur_pos[1]+2]
    elif (evt.GetKeyCode() == wx.WXK_DOWN and
          p.cur_pos[1] > 1):
        p.cur_pos = [p.cur_pos[0], p.cur_pos[1]-2]
    elif evt.GetKeyCode() == wx.WXK_SHIFT:
        if p.cur_pos[0] % 2 == 0:
            p.cur_pos = [1, 1]
        else:
            p.cur_pos = [0, 0]
    elif evt.GetKeyCode() == wx.WXK_RETURN:
        for pos in p.b_pos + p.r_pos:
            if pos[0] == p.cur_pos[0] and pos[1] == p.cur_pos[1]:
                return
        obj1, obj2 = draw_octagon(p.cur_pos, color=p.cur_color,
                                  direction=p.cur_dir, alpha=1)
        p._last_obj = (obj1, obj2)
        ppp = tuple(p.cur_pos)
        p._last_move = ppp
        if p.cur_color == 'red':
            p.r_pos.append(ppp)
            p.cur_color = 'black'
        else:
            p.b_pos.append(ppp)
            p.cur_color = 'red'
        if check_finish():
            p.viewer.canvas.canvas.Unbind(wx.EVT_KEY_UP)
            return
    elif evt.GetKeyCode() == wx.WXK_TAB:
        if p._last_obj is not None:
            page = params.page
            figure = page._artists[0]
            figure.patches.remove(p._last_obj[0])
            figure.patches.remove(p._last_obj[1])
            if p._last_move in p.r_pos:
                p.r_pos.remove(p._last_move)
            if p._last_move in p.b_pos:
                p.b_pos.remove(p._last_move)
            if p.cur_color == 'red':
                p.cur_color = 'black'
            else:
                p.cur_color = 'red'

        p._last_obj = None
        p._last_move = None

    if p.cur_pos[0] % 2 == 0:
        if p.cur_color == 'black':
            p.cur_dir = 'h'
        else:
            p.cur_dir = 'v'
    else:
        if p.cur_color == 'black':
            p.cur_dir = 'v'
        else:
            p.cur_dir = 'h'
    draw_cursor()


def r_check_finish(st, pos, et, rt):
    v = False
    if st in pos:
        rt.append(st)
        if st in et:
            return True
        pos.remove(st)
        if st[1] % 2 == 1:
            if r_check_finish((st[0]+2, st[1]), pos[:], et, rt):
                v = True
            if r_check_finish((st[0]-2, st[1]), pos[:], et, rt):
                v = True
        else:
            if r_check_finish((st[0],   st[1]+2), pos[:], et, rt):
                v = True
            if r_check_finish((st[0],   st[1]-2), pos[:], et, rt):
                v = True
        if r_check_finish((st[0]+1,   st[1]+1), pos[:], et, rt):
            v = True
        if r_check_finish((st[0]-1,   st[1]+1), pos[:], et, rt):
            v = True
        if r_check_finish((st[0]+1,   st[1]-1), pos[:], et, rt):
            v = True
        if r_check_finish((st[0]-1,   st[1]-1), pos[:], et, rt):
            v = True
    return v


def b_check_finish(st, pos, et, rt):
    if st in pos:
        rt.append(st)
        if st in et:
            return True
        pos.remove(st)
        if st[1] % 2 == 0:
            if b_check_finish((st[0]+2, st[1]), pos[:], et, rt):
                return True
            if b_check_finish((st[0]-2, st[1]), pos[:], et, rt):
                return True
        else:
            if b_check_finish((st[0],   st[1]-2), pos[:], et, rt):
                return True
            if b_check_finish((st[0],   st[1]+2), pos[:], et, rt):
                return True
        if b_check_finish((st[0]+1,   st[1]+1), pos[:], et, rt):
            return True
        if b_check_finish((st[0]+1,   st[1]-1), pos[:], et, rt):
            return True
        if b_check_finish((st[0]-1,   st[1]+1), pos[:], et, rt):
            return True
        if b_check_finish((st[0]-1,   st[1]-1), pos[:], et, rt):
            return True

    return False


def highlight_route(rt):
    import time
    objs = [draw_hl_octagon(pos, color='cyan', alpha=0.5) for pos in rt]
    print(objs)
    page = params.page
    viewer = wx.GetApp().TopWindow.find_bookviewer(page.get_figbook())
    viewer.draw()
    for x in range(20):
        time.sleep(0.1)
        if x % 2 == 0:
            for o in objs:
                o.set_alpha(0.8)
            viewer.draw_all()
        else:
            for o in objs:
                o.set_alpha(0.5)
            viewer.draw_all()


def check_finish():
    p = globals()['params']
    for x in [(0, 0), (2, 0), (4, 0), (6, 0), (8, 0)]:
        rt = []
        if r_check_finish(x, p.r_pos[:], [(0, 8), (2, 8), (4, 8), (6, 8), (8, 8)], rt):
            print('red win')
            highlight_route(rt)
            return True
    for x in [(0, 0), (0, 2), (0, 4), (0, 6), (0, 8)]:
        rt = []
        if b_check_finish(x, p.b_pos[:], [(8, 0), (8, 2), (8, 4), (8, 6), (8, 8)], rt):
            print('black win')
            highlight_route(rt)
            return True


def new_game():
    from __main__ import ifig_app
    import ifigure.events
    book = ifig_app.proj.onAddBook()
    i_page = book.add_page()
    page = book.get_page(i_page)
    page.realize()
    page.get_artist(0).set_facecolor('yellow')
    ifigure.events.SendOpenBookEvent(book, w=ifig_app,
                                     viewer=Connections, useProcessEvent=True)

    p = Params()
    globals()['params'] = p
    p.page = page
    p.viewer = ifig_app.find_bookviewer(book)
    p.viewer.SetSize((300, 300))
    draw_field()
    obj1, obj2 = draw_octagon([0, 0], color='black', direction='v', alpha=0.4)
    p.cursors = [obj1, obj2]
    p.viewer.canvas.canvas.Bind(wx.EVT_KEY_UP, handle_keyevent)



from ifigure.mdsplus.baseworker import MDSBaseWorker


class MDSProxyWorker(MDSBaseWorker):
    connection = None

    def __init__(self, *args, **kargs):
        super(MDSProxyWorker, self).__init__(*args, **kargs)
        pass

    def run(self, *args, **kargs):
        proc_name = self.name
        while True:
            job = self.task_queue.get(True)
            if job is None:
                self.task_queue.task_done()
                return
        return



import os
import ifigure


def base_dir():
    path = os.path.dirname(ifigure.__path__[0])
    base = os.path.dirname(path)
    return base


def bin_dir():
    return os.path.join(base_dir(), 'bin')



import wx
from wx import TextCtrl
from ifigure.utils.edit_list import TextCtrlCopyPasteGeneric as TextCtrl

font_h = None
font_w = None
font = None


def set_default_font():
    size = 12
    globals()['font'] = wx.Font(pointSize=size, family=wx.DEFAULT,
                                style=wx.NORMAL,  weight=wx.NORMAL,
                                faceName='Consolas')
    globals()['font_label'] = wx.Font(pointSize=size, family=wx.DEFAULT,
                                      style=wx.NORMAL,  weight=wx.BOLD,
                                      faceName='Consolas')
    dc = wx.ScreenDC()
    dc.SetFont(font)
    w, h = dc.GetTextExtent('A')
    globals()['font_h'] = h*1.5
    globals()['font_w'] = w


def truncate_str(txt, l):
    l = int(l)
    if len(txt) < l:
        return txt
    if l > 8:
        return txt[:3]+'...'+txt[-(l-6):]
    elif l > 3:
        return txt[:l-3]+'...'
    elif l > 0:
        return txt[:l]
    else:
        return ''


class TextCtrlTrunc(TextCtrl):
    def __init__(self, *args, **kargs):
        set_default_font()
        TextCtrl.__init__(self, *args, **kargs)
        self.SetFont(font)
        self.Bind(wx.EVT_SET_FOCUS, self.onSetFocus)
        self.Bind(wx.EVT_KILL_FOCUS, self.onKillFocus)
        self.Bind(wx.EVT_SIZE, self.onSize)
        self._data = ''
        self._mode = 'trunc'

    def SetValue(self, value):
        self._data = value
        if self._mode != 'trunc':
            return TextCtrl.SetValue(self, value)
        w, h = self.GetClientSize()
        if self.HasFocus():
            return TextCtrl.SetValue(self, value)
        else:
            value = truncate_str(value, w/font_w)
            TextCtrl.SetValue(self, value)

#    def GetValue(self):
#        return self._data

    def onSetFocus(self, evt):
        #        print 'set focus'
        self.SetValue(self._data)
        evt.Skip()

    def onKillFocus(self, evt):
        #        print 'kill focus'
        self._mode = 'trunc'
        evt.Skip()
        wx.CallAfter(self.SetValue, self._data)

    def onSize(self, evt):
        #        print 'kill focus'
        self.SetValue(self._data)
        evt.Skip()



from __future__ import print_function
import traceback
import os
import ifigure
import numpy as np
import ifigure.utils.cbook as cbook
from ifigure.mto.py_code import PyData
from ifigure.mdsplus.fig_mds import FigMds
from ifigure.mto.figobj_param import FigobjParam, FigobjData
import ifigure.utils.debug as debug
debug.debug_default_level = 1
dprint1, dprint2, dprint3 = debug.init_dprints('FigMdsData')


class FigMdsData(FigMds):
    def __init__(self, *args, **kargs):
        FigMds.__init__(self, *args, **kargs)
        self._plot_type = 'noplot'

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx = cbook.LoadImageFile(path, 'data.png')
        return [idx]

    @classmethod
    def property_in_palette(self):
        return ["mdssource"]

    def get_artist_extent(self, a=None):
        return [None]*4

    def generate_artist(self, *args, **kargs):
        pass

    def del_artist(self, *args, **kargs):
        pass

    def postprocess_data(self, ana, viewer,
                         color_order=['red', 'blue', 'yellow'],
                         ):
        ishot = ana.uishot
        if ishot < 0:
            return
        if ishot >= self.num_child():
            obj = FigobjData()
            self.add_child('data'+str(ishot+1), obj)
        else:
            obj = self.get_child(ishot)

        if self.get_script_local:
            txt = self.make_script(ana.shot)
            if txt is not None:
                try:
                    code = compile(txt, '<string>', 'exec')
                    g = {}
                    l = {}
                    exec(code, viewer.g, ana.result)
                except:
                    dprint1('error occured when processing data by script')
                    print('error occured when processing data by following script')
                    print('#####')
                    print(txt)
                    print('#####')
                    print(traceback.format_exc())
                    self.get_child(ishot).set_suppress(True)
                    ana.postprocess_done = True
                    return
        obj.setvar('shot', ana.shot)
        obj.setvar('global_data',
                   {key: ana.result[key] for key in ana.result
                    if (isinstance(ana.result[key], np.ndarray) or isinstance(ana.result[key], np.number))})
        ana.postprocess_done = True

    def process_title(self, ana):
        pass

    def call_refresh_artist(self, ishot):
        pass

    def active_plots(self, l):
        pass

    def set_mdsrange(self, value, a):
        self._default_xyrange = value

    def get_mdsrange(self, a):
        return self._default_xyrange

    def get_mdsfiguretype(self, a):
        return self._plot_type

    def get_script_local(self, a=None):
        v = self.get_figbook().getvar('mdsscript_main')
        if v is not None:
            return v
        return True

    def onDataSetting(self, evt, noapply=False):
        if self._session_editor is not None:
            try:
                self._session_editor.Raise()
                return
            except:
                pass
        from ifigure.mdsplus.dlg_mdssession_data import DlgMdsSessionData
        p = evt.GetEventObject().GetTopLevelParent()
        data = self.getvar('mdsvars')
        self._session_editor = DlgMdsSessionData(p, data=data,
                                                 figmds=self,
                                                 noapply=noapply)
        evt.Skip()

    def save_data2(self, data=None):
        if self._save_mode == 1:
            d = self.getvar('global_data')
            self.setvar('global_data', None)
            data = super(FigMdsData, self).save_data2(data)
            self.setvar('global_data', d)
        else:
            data = super(FigMdsData, self).save_data2(data)
        return data



from __future__ import print_function
#
#    Events for ifigure/treedict object
#
# *******************************************
#     Copyright(c) 2012- S.Shiraiwa
# *******************************************
#
#    2013.11.  post_process_event is added
#              I am little concerned if posting
#              event from thread is safe or not.
#
#              posting event through this routine
#              should be thread safe....?
import time
import os
import threading
from ifigure.widgets.book_viewer import BookViewer
import wx
from ifigure.utils.wx3to4 import deref_proxy
from weakref import ProxyType
import traceback
changed_in_queue = 0


EditFile = wx.NewEventType()
TD_EVT_EDITFILE = wx.PyEventBinder(EditFile, 1)
CloseFile = wx.NewEventType()
TD_EVT_CLOSEFILE = wx.PyEventBinder(CloseFile, 1)
Changed = wx.NewEventType()
TD_EVT_CHANGED = wx.PyEventBinder(Changed, 1)
Var0Changed = wx.NewEventType()
TD_EVT_VAR0CHANGED = wx.PyEventBinder(Var0Changed, 1)
ArtistSelection = wx.NewEventType()
TD_EVT_ARTIST_SELECTION = wx.PyEventBinder(ArtistSelection, 1)
ArtistDragSelection = wx.NewEventType()
TD_EVT_ARTIST_DRAGSELECTION = wx.PyEventBinder(ArtistDragSelection, 1)
ArtistReplace = wx.NewEventType()
TD_EVT_ARTIST_REPLACE = wx.PyEventBinder(ArtistReplace, 1)
ThreadStart = wx.NewEventType()
TD_EVT_THREAD_START = wx.PyEventBinder(ThreadStart, 1)
ShowPage = wx.NewEventType()
TD_EVT_SHOWPAGE = wx.PyEventBinder(ShowPage, 1)
PageShown = wx.NewEventType()
TD_EVT_PAGESHOWN = wx.PyEventBinder(PageShown, 1)
RangeChanged = wx.NewEventType()
TD_EVT_RANGECHANGED = wx.PyEventBinder(RangeChanged, 1)
OpenBook = wx.NewEventType()
TD_EVT_OPENBOOK = wx.PyEventBinder(OpenBook, 1)
CloseBook = wx.NewEventType()
TD_EVT_CLOSEBOOK = wx.PyEventBinder(CloseBook, 1)
FileChanged = wx.NewEventType()
TD_EVT_FILECHANGED = wx.PyEventBinder(FileChanged, 1)
PV_DrawRequest = wx.NewEventType()
PV_EVT_DrawRequest = wx.PyEventBinder(PV_DrawRequest, 1)
PV_DeleteFigobj = wx.NewEventType()
PV_EVT_DeleteFigobj = wx.PyEventBinder(PV_DeleteFigobj, 1)
PV_AddFigobj = wx.NewEventType()
PV_EVT_AddFigobj = wx.PyEventBinder(PV_AddFigobj, 1)
CV_CanvasSelected = wx.NewEventType()
CV_EVT_CanvasSelected = wx.PyEventBinder(CV_CanvasSelected, 1)
InteractiveAxesSelection = wx.NewEventType()
TD_EVT_IAXESSELECTION = wx.PyEventBinder(InteractiveAxesSelection, 1)
NewHistory = wx.NewEventType()
TD_EVT_NEWHISTORY = wx.PyEventBinder(NewHistory, 1)
RemoteCommand = wx.NewEventType()
REMOTE_COMMAND = wx.PyEventBinder(RemoteCommand, 1)
EvtTDDict = wx.NewEventType()
TD_EVT = wx.PyEventBinder(EvtTDDict, 1)
# undo/redo event
EvtUndo = wx.NewEventType()
TD_EVT_UNDO = wx.PyEventBinder(EvtUndo, 1)
EvtRedo = wx.NewEventType()
TD_EVT_REDO = wx.PyEventBinder(EvtRedo, 1)
FileSystemChanged = wx.NewEventType()
TD_EVT_FILESYSTEMCHANGED = wx.PyEventBinder(FileSystemChanged, 1)
CANVAS_DrawRequest = wx.NewEventType()
CANVAS_EVT_DRAWREQUEST = wx.PyEventBinder(CANVAS_DrawRequest, 1)
CloseBookRequest = wx.NewEventType()
TD_EVT_CLOSEBOOKREQUEST = wx.PyEventBinder(CloseBookRequest, 1)
#WorkerStartRequest= wx.NewEventType()
#EVT_WORKER_START_REQUEST = wx.PyEventBinder(WorkerStartRequest, 1)


class TreeDictEvent(wx.PyCommandEvent):
    """
    event for treedict to request edit a file
    """

    def __init__(self, evtType, id):
        wx.PyCommandEvent.__init__(self, evtType, id)
        self.dt = None
        self.event_name = ''
        self.time = time.time()
        self.BookViewerFrame_Processed = False

    def GetTreeDict(self):
        return self.dt

    def SetTreeDict(self, dt):
        self.dt = dt


def post_process_event(evt, handler, useProcessEvent=False):
    t = threading.current_thread()

    if t.name == 'MainThread':
        if not useProcessEvent:
            #             print 'posting event', evt.event_name
            wx.PostEvent(handler, evt)
        else:
            #             print 'processing event',evt.event_name
            handler.ProcessEvent(evt)
    else:
        if not useProcessEvent:
            wx.PostEvent(handler, evt)
        else:
            wx.CallAfter(handler.ProcessEvent, evt)


def SendSimpleTDEvent(evt, td=None, w=None, useProcessEvent=False, code='', handler=None):
    #    if td.get_root_parent().app is None: return

    if evt is None:
        evt = TreeDictEvent(EvtTDDict, wx.ID_ANY)
    evt.SetTreeDict(td)
    evt.SetEventObject(deref_proxy(w))
    evt.code = code
    if handler is None:
        handler = wx.GetApp().TopWindow.GetEventHandler()
    post_process_event(evt, handler, useProcessEvent)


def SendSimpleTDEvent2(evt, td=None, w=None, useProcessEvent=False):
    #    if td.get_root_parent().app is None: return

    evt = TreeDictEvent(EditFile, wx.ID_ANY)
    evt.SetTreeDict(td)
    evt.SetEventObject(deref_proxy(w))
    handler = wx.GetApp().TopWindow.GetEventHandler()
#    handler=td.get_root_parent().app.GetEventHandler()
    post_process_event(evt, handler, useProcessEvent)
#    if useProcessEvent:
#       handler.ProcessEvent(evt)
#    else:
#       wx.PostEvent(handler, evt)


def SendEditFileEvent(td, w=None, file=None, readonly=False):
    #    if td.get_root_parent().app is None: return
    if file is None: return

    handler = td.get_root_parent().app.GetEventHandler()
    evt = TreeDictEvent(EditFile, wx.ID_ANY)
    evt.SetTreeDict(td)
    evt.SetEventObject(deref_proxy(w))
    evt.file = file
    evt.readonly = readonly
    post_process_event(evt, handler, True)
#    handler.ProcessEvent(evt)


def SendCloseFileEvent(td, w=None, file=None):
    if td.get_root_parent().app is None:
        return
    if file is None: return
    
    handler = td.get_root_parent().app.GetEventHandler()
    evt = TreeDictEvent(CloseFile, wx.ID_ANY)
    evt.SetTreeDict(td)
    evt.SetEventObject(deref_proxy(w))
    evt.file = file
    post_process_event(evt, handler, True)
#    handler.ProcessEvent(evt)


def SendSelectionEvent(td, w=None, selections=[], multi_figobj=None, 
                       mode='replace', useProcessEvent=True):
    # events sent when figobj is selected
    # selections can be self.selection or self.axes_selection
    if td is None:
        return
    if td.get_root_parent().app is None:
        return

    evt = TreeDictEvent(ArtistSelection,
                        wx.ID_ANY)
    evt.selections = selections
    evt.SetTreeDict(td)
    evt.SetEventObject(deref_proxy(w))
    evt.event_name = 'selection'
    evt.mode = mode
    evt.multi_figobj=multi_figobj
    handler = td.get_root_parent().app.GetEventHandler()
    post_process_event(evt, handler, useProcessEvent)
#    handler.ProcessEvent(evt)

def SendDragSelectionEvent(td, w=None, selections=[], axes = None, useProcessEvent=True,
                           selected_index = None):
    # events sent when figobj is selected
    # selections can be self.selection or self.axes_selection
    if td is None:
        return
    if td.get_root_parent().app is None:
        return
    
    viewer = w.GetTopLevelParent()
    if viewer is None: return
    
    evt = TreeDictEvent(ArtistDragSelection, wx.ID_ANY)
    evt.selections = selections
    evt.SetTreeDict(td)
    evt.SetEventObject(deref_proxy(w))
    evt.event_name = 'multiselection'
    evt.mode = 'multi'
    evt.selected_index = selected_index

    handler = viewer.GetEventHandler()    
    post_process_event(evt, handler, useProcessEvent)

def SendReplaceEvent(td, w=None, a1=None, a2=None,
                     hl=False, a1all=None, a2all=None):
    # events sent when figobj replaces its artist
    # this event is processed immediately using
    # process event in order to keep data consisntency

    if td.get_root_parent().app is None:
        return

    app = td.get_root_parent().app
    book = td.get_root_figobj()
    viewer = app.find_bookviewer(book)
    if viewer is None:
        return

    evt = TreeDictEvent(ArtistReplace,
                        wx.ID_ANY)
    evt.a1 = a1
    evt.a2 = a2
    evt.hl = hl
    evt.a1all = a1all
    evt.a2all = a2all
    evt.SetTreeDict(td)
    if w is not None:
        evt.SetEventObject(deref_proxy(w))
    else:
        evt.SetEventObject(viewer)
    handler = viewer.GetEventHandler()
    post_process_event(evt, handler, True)
#    handler.ProcessEvent(evt)


def SendChangedEvent(td, w=None, useProcessEvent=False):
    # print 'sending changed event', td
    # traceback.print_stack()

    # I wanted to avoid issuing this event too many.
    # so it records if this event is already in event
    # queue.
    # However, if the event is not processed properly,
    # this approach could make a dead-lock.
    # Not running mainloop is one such case.
    # there may be a better way to do this...
    if not wx.GetApp().IsMainLoopRunning():
        return

    global changed_in_queue
    try:
        if td.get_root_parent().app is None:
            return
    except:
        return
#    if changed_in_queue: return
#    traceback.print_stack()
    evt = TreeDictEvent(Changed, wx.ID_ANY)
    evt.SetTreeDict(td)
    evt.SetEventObject(deref_proxy(w))

    handler = td.get_root_parent().app.GetEventHandler()
    post_process_event(evt, handler, useProcessEvent)
    changed_in_queue = changed_in_queue + 1


def SendVar0ChangedEvent(td, w=None):
    #    if td.get_root_parent().app is None: return
    if not hasattr(td.get_root_parent(), 'app'):
        return
    if td.get_root_parent().app is None:
        return

    evt = TreeDictEvent(Var0Changed, wx.ID_ANY)
    evt.SetTreeDict(td)
    evt.SetEventObject(deref_proxy(w))

    handler = td.get_root_parent().app.GetEventHandler()
#    handler.ProcessEvent(evt)
    post_process_event(evt, handler, False)
#    wx.PostEvent(handler, evt)
#    wx.Yield()


def SendThreadStartEvent(td, w=None, thread=None, useProcessEvent=False):
    if td.get_root_parent().app is None:
        return

    handler = td.get_root_parent().app.GetEventHandler()
    evt = TreeDictEvent(ThreadStart, wx.ID_ANY)
    evt.thread = thread
    evt.SetTreeDict(td)
    evt.SetEventObject(deref_proxy(w))
    post_process_event(evt, handler, useProcessEvent)


def SendFileChangedEvent(td, w=None, operation=None, param=None):
    '''
    opeartion 
    1) rename  : param = { "oldname": xxx , "newname": 
    2) delete  : param = { "oldname": xxx }
    '''

    if td.get_root_parent().app is None:
        return

    handler = td.get_root_parent().app.GetEventHandler()
    evt = TreeDictEvent(FileChanged, wx.ID_ANY)
    evt.op = operation
    evt.param = param
    evt.SetTreeDict(td)
    evt.SetEventObject(deref_proxy(w))
    post_process_event(evt, handler, False)
#    wx.PostEvent(handler, evt)


def SendShowPageEvent(td, w=None, inc='0'):
    try:
        app = td.get_root_parent().app
    except:
        return
    if app is None:
        return

    #inc = 1, -1, 'last', 'first'
    evt = TreeDictEvent(ShowPage, wx.ID_ANY)
    evt.SetTreeDict(td)
    if w is None:
        book = td.get_figbook()
        w = app.find_bookviewer(book).canvas
    evt.SetEventObject(deref_proxy(w))
    evt.inc = inc

    if w is None:
        handler = td.get_root_parent().app.GetEventHandler()
    else:
        handler = w
    evt.BookViewerFrame_Processed = False
    post_process_event(evt, handler, True)
#    handler.ProcessEvent(evt)


def SendRangeChangedEvent(td, w=None, name=''):
    evt = TreeDictEvent(RangeChanged, wx.ID_ANY)
    evt.name = name
    SendSimpleTDEvent(evt, td=td, w=w)
    print('send range change event')


def SendOpenBookEvent(td, w=None, viewer=BookViewer, param=None,
                      useProcessEvent=False, ipage=0, size=None, pos=None,
                      isPropShown=None, **kwargs):
    evt = TreeDictEvent(OpenBook, wx.ID_ANY)
    evt.viewer = viewer
    evt.param = param
    evt.ipage = ipage
    evt.size = size
    evt.pos = pos
    evt.isPropShown = isPropShown
    evt.kwargs = kwargs
    SendSimpleTDEvent(evt, td=td, w=w, useProcessEvent=useProcessEvent)


def SendCloseBookEvent(td, w=None):
    evt = TreeDictEvent(CloseBook, wx.ID_ANY)
    SendSimpleTDEvent(evt, td=td, w=w, useProcessEvent=True)


def SendIdleEvent(td, w=None):
    evt = wx.IdleEvent()
    handler = td.get_root_parent().app.GetEventHandler()
    wx.PostEvent(handler, evt)


def SendNewHistoryEvent(td, w=None):
    evt = TreeDictEvent(NewHistory, wx.ID_ANY)
    evt.event_name = 'newhistory'
    SendSimpleTDEvent(evt, td=td, w=w, handler=w.GetTopLevelParent())


def SendPageShownEvent(td, w=None):
    evt = TreeDictEvent(PageShown, wx.ID_ANY)
    SendSimpleTDEvent(evt, td=td, w=w)


def SendPVDrawRequest(td, w=None, wait_idle=False, refresh_hl=False,
                      useProcessEvent=False, caller=''):
    evt = TreeDictEvent(PV_DrawRequest, wx.ID_ANY)
    evt.wait_idle = wait_idle
    evt.refresh_hl = refresh_hl
    evt.caller = caller
    SendSimpleTDEvent(evt, td=td, w=w, useProcessEvent=useProcessEvent)


def SendPVAddFigobj(td, w=None, useProcessEvent=False):
    evt = TreeDictEvent(PV_AddFigobj, wx.ID_ANY)
    SendSimpleTDEvent(evt, td=td, w=w, useProcessEvent=useProcessEvent)


def SendPVDeleteFigobj(td, w=None, useProcessEvent=False):
    evt = TreeDictEvent(PV_DeleteFigobj, wx.ID_ANY)
    SendSimpleTDEvent(evt, td=td, w=w, useProcessEvent=useProcessEvent)


def SendCanvasSelected(td, w=None, useProcessEvent=True):
    evt = TreeDictEvent(CV_CanvasSelected, wx.ID_ANY)
    SendSimpleTDEvent(evt, td=td, w=w, useProcessEvent=useProcessEvent)


def SendInteractiveAxesSelection(td, w=None, callback=None, figaxes=None, handler=None):
    evt = TreeDictEvent(InteractiveAxesSelection, wx.ID_ANY)
    evt.callback = callback
    evt.figaxes = figaxes
    SendSimpleTDEvent(evt, td=td, w=w, handler=handler)


def SendRemoteCommandEvent(td, command=''):
    evt = TreeDictEvent(RemoteCommand, wx.ID_ANY)
    evt.command = command
    SendSimpleTDEvent(evt, td=td)


def SendUndoEvent(td, w=None, name=''):
    evt = TreeDictEvent(EvtUndo, wx.ID_ANY)
    evt.name = name
    SendSimpleTDEvent(evt, td=td, w=w, useProcessEvent=True)


def SendRedoEvent(td, w=None, name=''):
    evt = TreeDictEvent(EvtRedo, wx.ID_ANY)
    evt.name = name
    SendSimpleTDEvent(evt, td=td, w=w, useProcessEvent=True)


def SendFileSystemChangedEvent(td, w=None, name='', reload=True):
    evt = TreeDictEvent(FileSystemChanged, wx.ID_ANY)
    evt.owndir = td.owndir()
    evt.reload = reload
    SendSimpleTDEvent(evt, td=td, w=w, useProcessEvent=False)


def SendCanvasDrawRequest(w, all=False, delay=0.0, refresh_hl=False):
    evt = TreeDictEvent(CANVAS_DrawRequest, wx.ID_ANY)
    evt.all = all
    evt.delay = delay
    evt.refresh_hl = refresh_hl
    SendSimpleTDEvent(evt, w=w, handler=w)


def SendCloseBookRequest(w):
    evt = TreeDictEvent(CloseBookRequest, wx.ID_ANY)
    handler = wx.GetApp().TopWindow
    SendSimpleTDEvent(evt, w=w, handler=handler)



from __future__ import print_function
import multiprocessing as mp
#import sys, os, wx, weakref
#sys.path = [os.path.dirname(os.path.dirname(__file__)),] + sys.path

if __name__ == '__main__':
    from ifigure.piscope import piscope
    piscope()



from __future__ import print_function
from .miniframe_with_windowlist import MiniFrameWithWindowList
import os
import wx
import ifigure
import numpy as np
import wx.lib.platebtn as platebtn
import wx.lib.buttons as buttons
import wx.lib.agw.buttonpanel as bp
import ifigure.utils.cbook as cbook
from ifigure.mto.fig_axes import FigInsetAxes
import ifigure.widgets.dialog as dialog
from .navibar2 import make_bitmap_with_bluebox, TaskBtnList, ButtonInfo
from ifigure.utils.wx3to4 import wxCursorFromImage

from ifigure.utils.wx3to4 import evt_GetPosition

btasks0 = [('plus_theta', 'shift_theta.png', 0, '+ polar angle (shift-key: reserse direction)'),
#           ('minus_theta',   'minus_theta.png', 0, '- polar angle'),
           ('plus_phi',   'shift_phi.png', 0, '+ azimuthal angle (shift-key: reserse direction)'),
#           ('minus_phi',  'minus_phi.png', 0, '- azimuthal angle'),
           ('plus_offset',   'shift_cp.png', 0, '+/- offset (shift-key: reserse direction)'),
           ('yz_plane',  'yz_plane.png', 0, 'YZ plane'),
           ('xz_plane',  'xz_plane.png', 0, 'XZ plane'),           
           ('xy_plane',  'xy_plane.png', 0, 'XY plane'),                      
           ('flip_sign',    'flip_cp_side.png', 0, 'flip side'), 
           ('reset_cp',    'reset.png', 0, 'reset cp setting'), ]

# step size
delta_z = 0.02
delta_a = 5.
# interval(must be integer)
dtime = 150

class CutPlaneBar(bp.ButtonPanel):
    def __init__(self, parent, id=-1, text='', *args, **kargs):
        self.container = kargs.pop('container')

        super(CutPlaneBar, self).__init__(parent, id,  text, *args, **kargs)
        # mode of action (''=select, 'pan', 'zoom', 'text'....)
        self.mode = ''
        self.ptype = ''   # palette type ('pmode', 'amode')
        self.rotmode = False

        self.three_d_bar = False
        self.p0 = self.make_button_group(self, btasks0)
        self.btasks0 = btasks0[:]
        self.refresh_button()
        self.Fit()
        self.Bind(wx.EVT_KEY_DOWN, self._onKeyDown)
        self.Bind(wx.EVT_KEY_UP, self._onKeyUp)
        self.Bind(wx.EVT_LEAVE_WINDOW, self.OnLeave)
        self.Bind(wx.EVT_ENTER_WINDOW, self.OnEnter)        
        self.Bind(wx.EVT_SET_FOCUS, self._onFocus)
        
        self.place_bottoms()
        self._mouse_inside = False
        self._shift_down = False

    def _onKeyDown(self, evt):
        if evt.GetKeyCode() == wx.WXK_SHIFT:
            self._shift_down = True
            if self._mouse_inside:
                return
            
        wx.PostEvent(self.GetParent(), evt)

    def _onKeyUp(self, evt):
        if evt.GetKeyCode() == wx.WXK_SHIFT:
            self._shift_down = False
            if self._mouse_inside:
                return
            
        wx.PostEvent(self.GetParent(), evt)

    def _onFocus(self, evt):
        wx.PostEvent(self.GetParent(), evt)        

    def make_button_group(self, parent, btasks):

        bts = TaskBtnList()
        for items in btasks:
            btask, icon, tg, hint = items[:4]
            if btask == '---':
                bts.append('---')
#              bts.AddSpacer(icon)
                continue
            from ifigure.ifigure_config import icondir
            path = os.path.join(icondir, '16x16', icon)
            if icon[-3:] == 'png':
                im = wx.Image(path, wx.BITMAP_TYPE_PNG)
                image = im.ConvertToBitmap()
#              im = im.ConvertToGreyscale()
#              im = im.ConvertToMono(0,0,0)
                if im.HasAlpha():
                    im.ConvertAlphaToMask()
                crs = wxCursorFromImage(im)
            if icon[-3:] == 'bmp':
                im = wx.Image(path, wx.BITMAP_TYPE_BMP)
                image = im.ConvertToBitmap()
                if im.HasAlpha():
                    im.ConvertAlphaToMask()
                crs = wxCursorFromImage(im)
            # image.SetSize((8,8))

            btnl = ButtonInfo(self, wx.ID_ANY, image)
            btnl.custom_cursor = crs
            btnl.btask = btask
            btnl.bitmap1 = image
            #btnl.bitmap2 = make_bitmap_with_bluebox(image)
            if hint != '':
                btnl.SetShortHelp(hint)
            #if tg == 1:
            #    btnl.SetKind('toggle')
            #if len(items) > 4:
            #    btnl.use_in_2d_menu = items[4]
            #if len(items) > 5:
            #    btnl.use_in_3d_menu = items[5]

            bts.append(btnl)
            #           def func(evt, btask=btask): return self.OnButton(evt, btask)
            #           parent.Bind(wx.EVT_BUTTON, func, btnl)
        return bts

    def OnLeftDown(self, evt):
        ret = self.HitTest(evt_GetPosition(evt))
        if ret[0] == wx.NOT_FOUND:
            return bp.ButtonPanel.OnLeftDown(self, evt)

        bp.ButtonPanel.OnLeftDown(self, evt)
        btask = self.allbinfo[ret[0]].btask

        self._mouse_inside = True        
        self.OnButtonDown(evt, btask)

        win = self.GetTopLevelParent()
        canvas = win.canvas
        ax = win.canvas.axes_selection()
        if ax is None: return
        figax = ax.figobj
        if not figax.get_3d(): return

        wx.CallLater(dtime, self.timed_event, btask, canvas, ax)
        evt.Skip()

    def OnLeftUp(self, evt):
        ret = self.HitTest(evt_GetPosition(evt))
        if ret[0] == wx.NOT_FOUND:
            return bp.ButtonPanel.OnLeftUp(self, evt)

        bp.ButtonPanel.OnLeftUp(self, evt)
        btask = self.allbinfo[ret[0]].btask

        self._mouse_inside = False
        self.OnButtonUp(evt, btask)
        evt.Skip()
        
    def OnLeave(self, evt):
        self._mouse_inside = False
        evt.Skip()
        
    def OnEnter(self, evt):
        self._mouse_inside = True
        evt.Skip()
        
    def timed_event(self, btask, canvas, ax):

        def get_angles(limit1):
            rr = np.sqrt(limit1[0]**2 + limit1[1]**2)
            theta = np.arctan(rr/(limit1[2]+1e-5))
            if theta < 0:
                theta = np.pi + theta
            phi = np.arctan2(limit1[1], limit1[0])
            return theta, phi
        
        def get_norm(theta, phi):
            if theta > np.pi:
                theta = theta - np.pi
            elif theta < 0:
                theta = np.pi + theta

            vec = [np.sin(theta)*np.cos(phi),
                   np.sin(theta)*np.sin(phi),
                   np.cos(theta)]
            return vec

        if btask == 'plus_offset':
            limit1 = ax._lighting["clip_limit1"]
            limit2 = ax._lighting['clip_limit2']

            xmin, xmax = ax.get_xlim3d()
            ymin, ymax = ax.get_ylim3d()
            zmin, zmax = ax.get_zlim3d()
            nn  = np.sum(limit1*np.array([xmax-xmin, ymax-ymin, zmax-zmin]))

            delta = nn*delta_z

            if self._shift_down:
                limit2[0] = limit2[0] + delta
            else:
                limit2[0] = limit2[0] - delta
            ax._lighting['clip_limit2'] = limit2

        elif btask == 'plus_phi':
            th, ph = get_angles(ax._lighting['clip_limit1'])
            if self._shift_down:
                ph = ph  - delta_a/180*np.pi
            else:
                ph = ph  + delta_a/180*np.pi
            vec = get_norm(th, ph)
            ax._lighting['clip_limit1'] = vec

        elif btask == 'plus_theta':
            th, ph = get_angles(ax._lighting['clip_limit1'])
            limit2 = ax._lighting['clip_limit2']            
            if self._shift_down:
                th = th  + delta_a/180*np.pi            
                if th > np.pi:
                    limit2[1] = -limit2[1]
            else:
                th = th  - delta_a/180*np.pi
                if th < 0:
                    limit2[1] = -limit2[1]
                
            ax._lighting['clip_limit2'] = limit2
            vec = get_norm(th, ph)
            ax._lighting['clip_limit1'] = vec

        else:
            pass
        
        figax = ax.figobj
        figax.set_bmp_update(False)
        wx.CallAfter(canvas.draw)
        if self._mouse_inside:            
            wx.CallLater(dtime, self.timed_event, btask, canvas, ax)
        
    def AddButtonOrS(self, b):
        if isinstance(b, bp.ButtonInfo):
            if not self.three_d_bar:
                if not b.use_in_2d_menu:
                    return
            if self.three_d_bar:
                if not b.use_in_3d_menu:
                    return
            self.AddButton(b)
            self.allbinfo.append(b)
        else:
            self.AddSeparator()

    def Clear(self):
        self.allbinfo = []
        self.Freeze()
        bp.ButtonPanel.Clear(self)

    '''
    def set_toggle(self, btask):
        for p in self.p0:
            if p.btask != btask:
                p.SetStatus('Normal')
                p.SetToggled(False)
            else:
                p.SetStatus('Toggled')
                p.SetToggled(True)

    def set_bitmap2(self, btask):
        for p in self.p0:
            if p.btask == btask:
                p.SetBitmap(p.bitmap2)
            else:
                p.SetBitmap(p.bitmap1)
    '''
    def refresh_button(self):

        self.Clear()
        for b in self.p0:
            self.AddButtonOrS(b)
        self.DoLayout()

#    def OnKeyUp(self, evt):
#       if evt.GetKeyCode() == wx.WXK_SHIFT:
#           if self.mode == 'zoom': self.SetZoomUpDown('Up')

    def OnButtonDown(self, evt, btask):
        pass

    def OnButtonUp(self, evt, btask):
        win = self.GetTopLevelParent()
        canvas = win.canvas
        ax = win.canvas.axes_selection()
        if ax is None: return

        figax = ax.figobj
        if not figax.get_3d(): return
        
        v = self.container
        if btask == 'yz_plane':
            ax._lighting['clip_limit1'] = [1, 0, 0]
            figax.set_bmp_update(False)
            wx.CallAfter(canvas.draw)
        elif btask == 'xy_plane':
            ax._lighting['clip_limit1'] = [0, 0, 1]
            figax.set_bmp_update(False)
            wx.CallAfter(canvas.draw)
        elif btask == 'xz_plane':
            ax._lighting['clip_limit1'] = [0, 1, 0]
            figax.set_bmp_update(False)
            wx.CallAfter(canvas.draw)
        elif btask == 'flip_sign':
            limit2 = ax._lighting['clip_limit2']
            limit2[1] = -limit2[1]
            ax._lighting['clip_limit2'] = limit2
            figax.set_bmp_update(False)
            wx.CallAfter(canvas.draw)
        elif btask == 'reset_cp':
            ax._lighting['clip_limit1'] = [1, 0, 0]
            ax._lighting['clip_limit2'] = [0., 1, 0]
            figax.set_bmp_update(False)
            wx.CallAfter(canvas.draw)
        else:
            
            pass   

    '''
    def reset_btn_toggle_bitmap(self):
        self.set_toggle('')
        self.set_bitmap2('')
        self.refresh_button()
    '''

    def place_bottoms(self):
        # self.Fit()
        psize = self.GetSize()
        csize = self.GetParent().GetSize()
        self.SetPosition((csize[0]-psize[0]-4,
                          4))

def add_cutplane_btns(parent):
    canvas = parent.canvas.canvas
    cutplanebtns = CutPlaneBar(canvas, wx.ID_ANY, container=parent)
    cutplanebtns.Show()

    return cutplanebtns



from __future__ import print_function
#
#  Name   : py_script
#
#          mto to hold a script in the tree
#
#  Author :
#         Syun'ichi Shiraiwa
#  E-mail :
#         shiraiwa@psfc.mit.edu
#
#
#  History : 0.5 sometime around Apr. 2012
#            0.6 stop() command
#            0.7 8/30/12 keyborad interupt
#
from ifigure.mto.treelink import TreeLink
import collections
import weakref
import os
import logging
import wx
import threading
import types
import shutil
import multiprocessing
import sys
import traceback
import ifigure.utils.pickle_wrapper as pickle
from ifigure.mto.py_code import PyCode
import ifigure.utils.cbook as cbook
import ifigure
import ifigure.events
import ifigure.widgets.dialog as dialog
from ifigure.mto.fileholder import FileHolder

import ifigure.utils.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('PyScript')

txt1 = '(job involving GUI access, such as graphics generation,\nshould not use this option.)'


class ScriptStop(Exception):
    def __init__(self, message=''):
        super(Exception, self).__init__()
        self.message = message

    def __str__(self):
        return repr(self.messge)


class ExitScript(Exception):
    def __init__(self, message=''):
        super(Exception, self).__init__()
        self.message = message

    def __str__(self):
        return repr(self.message)


def stop(message=''):
    raise ScriptStop(message)


def exit(message=''):
    raise ExitScript(message)


class AnsHolder(object):
    def get_ansfunc(self):
        self._ans = None

        def ans(value=None, obj=self):
            obj._ans = value
            raise ExitScript('')
        return ans

    def get_ansfunc_array(self, l):
        self._ans = [None]*l
        ret = [None]*l
        for k in range(l):
            def ans(value, obj=self, i=k):
                obj._ans[i] = value
                raise ExitScript('')
            ret[k] = ans
        return ret

    def RunA(self, *args, **kargs):
        self.Run(*args, **kargs)
        if 'ans' in kargs:
            obj = kargs['ans'].__defaults__[1]
        else:
            obj = self
        val = obj._ans
        obj._ans = None
        #val =  self._ans
        #self._ans = None
        return val

    def __call__(self, *args, **kargs):
        return self.RunA(*args, **kargs)

    @property
    def ans(self):
        return self._ans


class AbsScript(object):
    #
    #  interface to run python script
    #  this part should be independent from
    #  GUI and tree structure.
    #    provide
    #        automatic update check
    #        launch editor
    #
    def __init__(self):
        #        self._script_file=''
        self._script = ''
        self._script_co = None
        self._script_mtime = 0.
        self._debug = 0

    def is_objectold(self, obj):
        script_file = obj.path2fullpath()
        if script_file == '':
            return False
        if (os.path.getmtime(script_file) >
                self._script_mtime):
            return True
        return False

    def run_script(self, obj=None, top=None, param=None,
                   wdir=None, app=None, model=None,
                   **kargs):

        script_file = obj.path2fullpath()
        logging.basicConfig(level=logging.DEBUG)
        if script_file == '':
            return

        if (os.path.getmtime(script_file) >
                self._script_mtime):
            self.load_script(script_file)
        if self._script_co is not None:
            debug = self._debug

            def write_log(*args):
                top.write_log(obj, model, *args)

            lc = obj.get_namespace()
            lc2 = {"obj": obj, "proj": top, "top": top,  "param": param,
                   "wdir": wdir, "app": app, "model": model,  "stop": stop,
                   "exit": exit, "write_log": write_log}
            for key in lc2:
                lc[key] = lc2[key]
            for key in kargs:
                lc[key] = kargs[key]

            self._debug = 0  # make sure that next time will run in normal mode
            try:
                if debug == 0:
                    exec(self._script_co, lc, lc)
                elif debug == 1:
                    import pdb
                    pdb.run(self._script_co, lc, lc)
                elif debug == 2:
                    app.script_editor.RunSED(
                        self._script_co, lc, lc, script_file)

            except ExitScript:
                return True
            except ScriptStop as e:
                print(('Script execution stops : ', e.message))
                return False
            except Exception:
                print('script exectuion failed')
                print(traceback.format_exc())
                return False
                #logging.exception("Script Execution Failed")
            return True

    def load_script(self, file):
        if file is None:
            return
        if file == '':
            return
        try:
            a, b, c = cbook.LoadScriptFile(file)
            self._script = a
            self._script_co = b
            self._script_mtime = c
        except:
            self._script = ''
            self._script_co = None
            self._script_mtime = -1

            print('Failed to compile script '+file)
            print(traceback.format_exc())
            return traceback.format_exc()
        return ''

#  this decorator
class doc_decorator(object):
    def __init__(self, func):
        self.func = func

    def __call__(self, *args, **kargs):
        return self.func.__call__(self.obj, *args, **kargs)

    def __get__(self, instance, owner):
        self.cls = owner
        self.obj = instance
        self.__call__.__func__.__doc__ = instance._run_doc
        return self.__call__


class PyScriptLink(TreeLink, AnsHolder):
    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx = cbook.LoadImageFile(path, 'script_link.png')
        return [idx]

    def __call__(self, *args, **kargs):
        model = self.get_pymodel()
        kargs['model'] = model
        kargs['ans'] = self.get_ansfunc()

        if self.is_linkalive:
            t = self.get_linkobj()
            return t.__call__(*args, **kargs)

    def Run(self, *args, **kargs):
        model = self.get_pymodel()
        kargs['model'] = model
        kargs['ans'] = self.get_ansfunc()

        if self.is_linkalive:
            t = self.get_linkobj()
            return t.Run(*args, **kargs)

    def Debug(self, *args, **kargs):
        model = self.get_pymodel()
        kargs['model'] = model
        kargs['ans'] = self.get_ansfunc()

        if self.is_linkalive:
            t = self.get_linkobj()
            return t.Debug(*args, **kargs)

    def do_run(self, *args, **kargs):
        return self.Run(*args, **kargs)

    def do_run_f(self, *args, **kargs):
        model = self.get_pymodel()
        kargs['model'] = model
        kargs['ans'] = self.get_ansfunc()

        if self.is_linkalive:
            t = self.get_linkobj()
            return t.do_run_f(*args, **kargs)

    def do_run_t(self, e, *args, **kargs):
        model = self.get_pymodel()
        kargs['model'] = model
        kargs['ans'] = self.get_ansfunc()

        if self.is_linkalive:
            t = self.get_linkobj()
            return t.do_run_t(e, *args, **kargs)

    def tree_viewer_menu(self):
        return [('+Script', None, None),
                ('Run (F9)', self.onRunQuick, None),
                ('Run...', self.onRunFull, None),
                ('Edit...', self.onEditScript, None),
                ('!', None, None),
                ('---', None, None)] + \
            super(TreeLink, self).tree_viewer_menu()

    def onRunQuick(self, e):
        if not self.is_linkalive:
            return
        t = self.get_linkobj()

        ns = self.get_root_parent().app.shell.lvar.copy()
        args = t.getvar('script_args')
        kargs = t.getvar('scriptk_kwargs')
        if kargs is None:
            kargs = '{}'
        if args is None:
            args = '()'
        try:
            args = eval(args, ns, ns)
        except Exception:
            args = ()
        kargs = t.getvar('scriptk_kwargs')
        try:
            kargs = eval(kargs, ns, ns)
        except Exception:
            kargs = {}
        self.do_run_f(*args, **kargs)

    def onRunFull(self, e):
        if not self.is_linkalive:
            return
        t = self.get_linkobj()

        args = t.getvar('script_args')
        if args is None:
            args = '()'
        kargs = t.getvar('scriptk_kwargs')
        if kargs is None:
            kargs = '{}'
        shellvar = self.get_root_parent().app.shell.lvar

        # the following makes a crash
        list = [["", "Run script ("+self.compact_name()+")", 2],
                ["Args", str(args), 100, {'ns': shellvar}],
                ["Kwargs", str(kargs), 100, {'ns': shellvar}],
                #                ["model variable", self.getvar('script_model'), 0, {}],
                ["Use thread", False, 3,
                 {"text": txt1, "expand": True}]]

        self.get_root_parent().app.proj_tree_viewer.OpenPanel(list, self, 'handle_run_full',
                                                              event=e)
        e.Skip()

    def handle_run_full(self, value, event=None):
        if not self.is_linkalive:
            return

        t = self.get_linkobj()
        t.setvar('script_args', value[1][0])
        t.setvar('script_kwargs', value[2][0])

        kargs = value[2][1]
        kargs['model'] = self.get_pymodel()
        if value[3]:
            wx.CallAfter(self.do_run_t, event, *(value[1][1]), **kargs)
        else:
            wx.CallAfter(self.do_run_f,  *(value[1][1]), **kargs)

    def onEditScript(self, evt):
        if self.is_linkalive:
            t = self.get_linkobj()
            t.onEditScript(evt)

    def init_after_load(self, olist, nlist):
        xxx = super(PyScriptLink, self).init_after_load(olist, nlist)
        self._status = ''
        return xxx


class PyScript(PyCode, FileHolder, AnsHolder):
    def __init__(self, parent=None, script='', src=None):
        self._script = AbsScript()
        self._simplemenu = False
        super(PyScript, self).__init__(parent=parent, src=src)
 #       self.setvar("script file", None)
        self.setvar("background", False)
        self.setvar("script_args", '()')
        self.setvar("script_kwargs", '{}')
        self.setvar("script_model", '')
        self._script.load_script(None)
        self._can_have_child = False
        self._has_private_owndir = False
        self._run_doc = ''
        self._save_link = False

    @classmethod
    def isPyScript(self):
        return True

    @classmethod
    def get_namebase(self):
        return 'script'

    def classimage(self):
        if PyScript._image_load_done is False:
            PyScript._image_id = self.load_classimage()
            PyScript._image_load_done = True

        if self.getvar('pathmode') == 'owndir':
            return PyScript._image_id[0]
        else:
            return PyScript._image_id[1]


    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = cbook.LoadImageFile(path, 'script.png')
        idx2 = cbook.LoadImageFile(path, 'script_ext.png')
        return [idx1, idx2]

    @classmethod
    def can_have_child(self, child=None):
        return False

    def pv_kshortcut(self, key):
        '''
        define keyborad short cut on project viewer
        must return a method responding key
        '''
        if key == wx.WXK_F9:
            return self.onRunQuick
        if key == wx.WXK_RETURN:
            return self.onEditScript
        return None

    def on_runf_menu(self, evt):
        return self.onRunF()

    @doc_decorator
    def Run(self, *args, **kargs):
        if self.getvar("background") and self._script._debug == 0:
            return self.do_run_t(None, *args,   **kargs)
        else:
            return self.do_run_f(*args, **kargs)

    def RunT(self, *args, **kargs):
        return self.do_run_t(None, *args,   **kargs)

#    def Debug1(self,  *args, **kargs):
#        self._script._debug = 1
#        wx.CallAfter(self.Run, *args, **kargs)

    def Debug(self,  *args, **kargs):
        se = wx.GetApp().TopWindow.script_editor
        if not se.CheckDebuggerStatus():
            ret = dialog.message(
                self, 'Debugger is running.', 'Please wait', 0)
            return
        self._script._debug = 2
        se.GoDebugMode(self.path2fullpath(), enter=True)
        wx.CallAfter(self.Run, *args, **kargs)

    def do_run(self, *args, **kargs):
        return self.Run(*args, **kargs)

    def do_run_f(self, *args, **kargs):

        if 'model' in kargs:
            model = kargs['model']
            ans = kargs['ans']
            del kargs['model']
            del kargs['ans']
        else:
            model = self.get_pymodel()
            ans = self.get_ansfunc()
        p = model
        param = None
        while p is not None:
            if p.has_child('param'):
                param = p.param
                break
            p = p.get_parent()
        if param is None:
            if self.get_root_parent().has_child('parameters'):
                param = self.get_root_parent().has_child('parameters')

        kargs = {"obj": self,
                 "top": self.get_root_parent(),
                 "app": self.get_root_parent().app,
                 "wdir": self.get_root_parent().getvar("wdir"),
                 "model": model,
                 "param": param,
                 "ans": ans,
                 "args": args,
                 "kwargs": kargs}

        if self._suppress is False:
            val = self._script.run_script(**kargs)
            ifigure.events.SendChangedEvent(self)
            return val
        else:
            dprint1('script is suppressed')

    def do_run_t(self, e, *args, **kargs):
        app = wx.GetApp().TopWindow
        maxt = app.aconfig.setting['max_thread']
#        print len(app.logw.threadlist), maxt
        if len(app.logw.threadlist) < maxt:
            #aargs = [model] + list(args)
            t = threading.Thread(target=self.do_run_f, args=args, kwargs=kargs)
            self._status = 'launching...'
            # useProcessEvent is True so that threadlist in logw is
            # updated
            ifigure.events.SendThreadStartEvent(self, w=self.get_app(),
                                                thread=t,
                                                useProcessEvent=True)
        else:
            wx.CallLater(1000, self.do_run_t, e, *args, **kargs)
        if e is not None:
            e.Skip()

    def call(self, name, *args, **kargs):
        '''
        call function in script file
             call(fname, *args, **kargs)
        note:
           this is a hacky program and use of this
           method should be restricted to write
           something temporary. For more permanent 
           implementation, put function in py_module
           or use an usual "import....". 

        restriction:
           default arguments can not be used.
        '''
        f = self.get_function(name)
        if f is None:
            return
        return f(*args, **kargs)

    def get_function(self, name):
        '''
        call function in script file
             call(fname, *args, **kargs)
        note:
           this is a hacky program and use of this
           method should be restricted to write
           something temporary. For more permanent 
           implementation, put function in py_module
           or use an usual "import....". 

        restriction:
           default arguments can not be used.
        '''
        code = None
        for code in self._script._script_co.co_consts:
            if isinstance(code, types.CodeType):
                if code.co_name is name:
                    break

        if code is None:
            raise AttributeError("Function does not exist:  "+name)
        else:
            #
            #            import types
            #
            def hack():
                pass
            hack.__code__ = code
            hack.__name__ = name
            return hack

    def onRunQuick(self, e):
        ns = self.get_root_parent().app.shell.lvar.copy()
        args = self.getvar('script_args')
        kargs = self.getvar('scriptk_kwargs')
        if kargs is None:
            kargs = '{}'
        if args is None:
            args = '()'
        try:
            args = eval(args, ns, ns)
        except Exception:
            args = ()
        kargs = self.getvar('scriptk_kwargs')
        try:
            kargs = eval(kargs, ns, ns)
        except Exception:
            kargs = {}
        self.do_run_f(*args, **kargs)

    def onRunFull(self, e, title='Run script'):
        args = self.getvar('script_args')
        if args is None:
            args = '()'
        kargs = self.getvar('scriptk_kwargs')
        if kargs is None:
            kargs = '{}'
#        self.do_run_t(e, *args, **kargs)
#        return
        shellvar = self.get_root_parent().app.shell.lvar
        # the following makes a crash
        list = [["", "Run script ("+self.compact_name()+")", 2],
                ["Args", str(args), 100, {'ns': shellvar}],
                ["Kwargs", str(kargs), 100, {'ns': shellvar}],
                ["model variable", self.getvar('script_model'), 0, {}],
                ["Use thread", False, 3,
                 {"text": txt1, "expand": True}]]

        self.get_root_parent().app.proj_tree_viewer.OpenPanel(list, self, 'handle_run_full',
                                                              title=title, event=e)
        e.Skip()

    def handle_run_full(self, value, event=None):
        self.setvar('script_args', value[1][0])
        self.setvar('script_kwargs', value[2][0])
        self.setvar('script_model', str(value[3]))
        if str(value[3]).strip() == '':
            models = [self.get_pymodel()]
        else:
            models = self.expand_path(str(value[3]))

        if len(models) == 1:
            anses = [self.get_ansfunc()]
        else:
            anses = self.get_ansfunc_array(len(models))
        for ans, m in zip(anses, models):
            kargs = value[2][1]
            kargs['model'] = m
            kargs['ans'] = ans

            if value[4] and self._script._debug == 0:
                wx.CallAfter(self.do_run_t, event, *(value[1][1]), **kargs)
            else:
                wx.CallAfter(self.do_run_f,  *(value[1][1]), **kargs)

    def tree_viewer_menu(self):
     # return MenuString, Handler, MenuImage
        flag = ''
        if self._script is not None:
            d = self._script._debug
            try:
                if self._script.is_objectold(self):
                    file = self.path2fullpath()
                    self.load_script(file)
            except:
                dprint1('Script loading failed ' + str(self))
                flag = '*'
        else:
            d = 0
        a = ['0', '1', '2']
        a[d] = '-'+a[d]

        se = wx.GetApp().TopWindow.script_editor
        dflag = '' if se.CheckDebuggerStatus() else '-'

        if self._simplemenu:
            return [('+Script', None, None),
                    (flag+'Run (F9)', self.onRunQuick, None),
                    (flag+'Run...', self.onRunFull, None),
                    (flag+'Edit...', self.onEditScript, None),
                    ('Import File...', self.onImportScriptFile, None),
                    (flag+'Export File...', self.onExportScriptFile, None),
                    (dflag+'Debug...', self.onDebug2, None),
                    ('!', None, None),
                    ('---', None, None)] + \
                super(PyCode, self).tree_viewer_menu()
        else:
            m = [('+Script', None, None),
                 ('Run (F9)', self.onRunQuick, None),
                 ('Run...', self.onRunFull, None), ]
            if not self._is_in_pysol():
                m.append((flag+'Run at Sol...', self.onRunAtSol, None))
            m.extend([(flag+'Edit...', self.onEditScript, None),
                      ('Import File...', self.onImportScriptFile, None),
                      (flag+'Export File...', self.onExportScriptFile, None),
                      (flag+'Update scripts in Sol...', self.onExportToSol, None),
                      #               ('+Degug...', None, None),
                      #               ('Run in PDB', self.onDebug1, None),
                      (dflag+'Debug...', self.onDebug2, None),
                      #               ('Import File', self.onImportScriptFile, None),
                      #               ('!', None, None),
                      #               ('+Run Mode', None, None),
                      #               ('-'*self.is_manualonly()+'Manual Only', self.onManual, None),
                      #               ('-'*(not self.is_manualonly())+'Auto', self.onAuto, None),
                      #               ('!', None, None),
                      ('!', None, None),
                      ('---', None, None)])
            m = m + super(PyScript, self).tree_viewer_menu()
            return m

#    def onDebug1(self, e):
#        self._script._debug = 1
#        self.onRunQuick(e)

    def onDebug2(self, e):
        se = wx.GetApp().TopWindow.script_editor
        if not se.CheckDebuggerStatus():
            ret = dialog.message(
                self, 'Debugger is running.', 'Please wait', 0)
            return
        self._script._debug = 2
        se.GoDebugMode(self.path2fullpath(), enter=True)
        self.onRunFull(e, title='Debug run...')

    def onManual(self, e):
        self._manual_only = True

    def onAuto(self, e):
        self._manual_only = False

    def _is_in_pysol(self):
        from ifigure.mto.py_code import PySol
        p = self
        while p is not None:
            if isinstance(p, PySol):
                return True
            p = p.get_parent()
        return False

    def _make_list1(self, sol_list, p):
        from ifigure.mto.py_code import PySol
        for name, child in p.get_children():
            if isinstance(child, PySol):
                src = child.getvar('source_model')
                if src is not None:
                    src = self.find_by_full_path(src)
                    if src is not None:
                        if src.is_descendant(self):
                            sol_list.append(child.get_full_path())
                self._make_list1(sol_list, child)
        return sol_list

    def _make_list2(self, sol_list, p):
        from ifigure.mto.py_code import PySol
        if isinstance(p, PySol):
            src_name = p.getvar('source_model')
            if src_name is not None:
                src = self.find_by_full_path(src_name)
                if src is not None:
                    if src.is_descendant(self):
                        sol_list.append((p, src_name))
        for name, child in p.get_children():
            sol_list = self._make_list2(sol_list, child)
        return sol_list

    def _ask_at_sol_conditionl(self, e):
        from ifigure.mto.py_code import PySol
        proj = self.get_root_parent()
        sol_list = self._make_list1([], proj)

        if len(sol_list) == 0:
            dlg = wx.MessageDialog(self.get_app(),
                                   "No solution based on this model is found",
                                   "Error", wx.OK | wx.ICON_INFORMATION)
            ans = dlg.ShowModal()
            dlg.Destroy()
            return

        shellvar = proj.app.shell.lvar
        list = [[None, "Run at SOL ("+self.compact_name()+")", 2],
                ["sol", sol_list[0], 4, {'style': wx.CB_DROPDOWN,
                                         'choices': sol_list}],
                ["Args", '()', 100, {'ns': shellvar}],
                ["Kwargs", '{}', 100, {'ns': shellvar}],
                ["Use thread", False, 3,
                 {"text": txt1, "expand": True}]]
        return list

    def _find_script_in_sol(self, sol, src_name):
        model = self.find_by_full_path(src_name)
#        path =  [model.name]+ model.get_relative_path(self)
        path = [model.name] + model.get_td_path(self)[1:]
        obj = sol
        for name in path[:-1]:
            obj = obj.get_child(name=name)
        script = obj.get_child(name=path[-1])
        if script is None:
            if not obj.has_owndir():
                obj.mk_owndir()
            script = PyScript()
            obj.add_child(path[-1], script)
            fpath = os.path.join(script.owndir(), path[-1]+'.py')
            script.set_path_pathmode(fpath)
            print(('creating...', script))
            ifigure.events.SendChangedEvent(script, w=wx.GetApp().TopWindow)
        return script

    def onRunAtSol(self, e, title='Run at sol'):
        self._script._debug = 0
        list = self._ask_at_sol_conditionl(e)
        app = wx.GetApp().TopWindow
        app.proj_tree_viewer.OpenPanel(
            list, self, 'handle_run_at_sol', title=title)

    def handle_run_at_sol(self, value):
        proj = self.get_root_parent()
        base_sol = str(value[1])
        sol_list = self._make_list2([], proj.find_by_full_path(base_sol))
        for sol, src_name in sol_list:
            script = self._find_script_in_sol(sol, src_name)
            print(('running...', script.get_full_path()))  # value
            if value[4]:
                wx.CallAfter(script.do_run_t, None, *
                             (value[2][1]), **(value[3][1]))
            else:
                #                script.do_run_f(*(value[2][1]), **(value[3][1]))
                wx.CallAfter(script.do_run_f,  *(value[2][1]), **(value[3][1]))

    def _select_sol(self, e):
        from ifigure.mto.py_code import PySol
        proj = self.get_root_parent()
        sol_list = self._make_list1([], proj)

        if len(sol_list) == 0:
            dlg = wx.MessageDialog(self.get_app(),
                                   "No solution based on this model is found",
                                   "Error", wx.OK | wx.ICON_INFORMATION)
            ans = dlg.ShowModal()
            dlg.Destroy()
            return

        shellvar = proj.app.shell.lvar
        list = [["", "Select Sol", 2],
                [" ", sol_list[0], 4, {'style': wx.CB_DROPDOWN,
                                       'choices': sol_list}], ]
        return list

    def onExportToSol(self, e):
        list = self._select_sol(e)
        app = wx.GetApp().TopWindow
        app.proj_tree_viewer.OpenPanel(list, self, 'handle_export_to_sol')

    def handle_export_to_sol(self, value):
        proj = self.get_root_parent()
        base_sol = str(value[1])
        sol_list = self._make_list2([], proj.find_by_full_path(base_sol))
        src_file = FileHolder.path2fullpath(self)
        for sol, src_name in sol_list:
            script = self._find_script_in_sol(sol, src_name)
            print(('updating...', script))
            dest_file = FileHolder.path2fullpath(script)
            shutil.copyfile(src_file, dest_file)

#    def run_all_sol(self, sol, src_name, bg, *args, **kargs):

    def pv_kshortcut(self, key):
        '''
        define keyborad short cut on project viewer
        must return a method responding key

        '''
#        if not self._simplemenu:

        if key == wx.WXK_F9:
            return self.onRunQuick
        return super(PyScript, self).pv_kshortcut(key)

#        return None
#    def onRename(self, evt):
#        super(PyScript, self).onRename(evt)

    def onEditScript(self, e):
        fpath = FileHolder.path2fullpath(self)
        if fpath == '':
            return
        self._script.load_script(fpath)
#        app = self.get_root_parent().app
#        if app.helper.setting['use_editor']:
#            txt = app.helper.setting['editor'].format(fpath)
#            print txt
#            os.system(txt)
# else:
        from ifigure.mto.treedict import str_td
        fpath = str_td(fpath)
        fpath.td = self.get_full_path()
        ifigure.events.SendEditFileEvent(self,
                                         w=e.GetEventObject(),
                                         file=fpath)

    def onImportScriptFile(self, e=None, file=None):
        from ifigure.utils.addon_utils import onLoadFile
        ret = onLoadFile(self, message="Enter script file name",
                         modename='pathmode',
                         pathname='path',
                         extname='.py',
                         reject_loc=['wdir', 'owndir'],
                         file=file,
                         wildcard='*.py')

        if not ret:
            return
        file = self.path2fullpath()
        if self.getvar('pathmode') == 'owndir':
            newname = str(os.path.basename(file).split('.')[0])
            self.rename(newname)
        else:
            self.load_script(file)
#        ifigure.events.SendFileSystemChangedEvent(self)
#        self.load_script(file)
        return
        file = dialog.read(None, message="Select script",
                           wildcard='*.py')
        mode, path = FileHolder.fullpath2path(self, file)
        if (mode == 'wdir' or mode == 'owndir'):
            m = 'Import should import from somewhere outside project directory',
            ret = dialog.message(None, message=m,
                                 title='Import error')
            return
        self.import_script(file)

    def onExportScriptFile(self, e):
        opath = self.path2fullpath()
        file = dialog.write(None, defaultfile=opath,
                            message="Enter script file name",
                            wildcard='*.py')
        if file == '':
            return
        mode, path = FileHolder.fullpath2path(self, file)
        if (mode == 'wdir' or mode == 'owndir'):
            m = 'Export should export to somewhere outside project directory',
            ret = dialog.message(None, message=m,
                                 title='Import error')
            return
        path = FileHolder.path2fullpath(self)
        shutil.copyfile(path, file)

    def export_script(self, file):
        '''
        export scirpt file to file.
        the destination should be outside
        of wdir
        '''
        path = self.path2fullpath()
        mode, path0 = self.fullpath2path(file)
        if mode == 'wdir':
            # do not write inside proj directory
            print('can not export to wdir')
            return
        if mode == 'owndir':
            # do not write inside proj directory
            print('can not export to owndir')
            return

        shutil.copyfile(path, file)

    def import_script(self, file, copy_file=True):
        '''
        import *.py file from outside of wdir
        the location of original file will be
        saved to "ofile_path", "ofile_pathmode"
        '''
        mode, path = self.fullpath2path(file)
        if mode == 'wdir':
            # do nothing in this case
            print('can not import from wdir')
            return
        if mode == 'owndir':
            # do not write inside proj directory
            print('can not import from owndir')
            return

        self.setvar("ofile_path", str(path))
        self.setvar("ofile_pathmode", str(mode))

        if copy_file:
            if not self.has_owndir():
                self.mk_owndir()
            file2 = os.path.join(self.owndir(), self.name+'.py')
            shutil.copyfile(file, file2)
            self.load_script(file2)
        else:
            self.remove_ownitem(items=['path'])
            self.set_path_pathmode(path,  mode,  'path', '.py',
                                   checklist=['proj', 'home', 'abs'])
            self.load_script(file)

    def set_breakpoint(self, line, set=True):
        from ifigure.widgets.debugger_core import add_breakpoint, rm_breakpoint
        from ifigure.widgets.debugger import check_debugger_instance

        check_debugger_instance()
        path = FileHolder.path2fullpath(self)
        if set:
            add_breakpoint(path, line)
        else:
            rm_breakpoint(path, line)

    def get_breakpoint(self):
        from ifigure.widgets.debugger_core import get_breakpoint
        from ifigure.widgets.debugger import check_debugger_instance

        check_debugger_instance()
        path = FileHolder.path2fullpath(self)
        return get_breakpoint(path)

    def load_script(self, file):
        self.set_path_pathmode(file)
        err = self._script.load_script(file)
        if self._script._script_co is not None:
            self.set_doc(file)
        return err

    def set_doc(self, file):
        fid = open(file, 'r')

        doc = []
        l = 0
        doc_exist = False
        for x in fid.readlines():
            if ((x.startswith('"""') and l == 0) or
                    (x.startswith("'''") and l == 0)):
                doc_exist = True
            if ((x.startswith('"""') and l != 0) or
                    (x.startswith("'''") and l != 0)):
                doc.append(x)
                break
            if not doc_exist:
                break
            doc.append(x)
            l = l + 1
        doc = ''.join(doc)
        self._run_doc = doc
        self.__doc__ = doc
        fid.close()

    def has_script(self):
        file = self.path2fullpath()
        if (file != '' and
                os.path.exists(file)):
            return True
        return False

    def init_after_load(self, olist, nlist):
        #        print "loading", self.getvar("module file")
        name = self._name
        file = self.path2fullpath()
#        print file
        if (file != '' and
                os.path.exists(file)):
            self.load_script(file)
        self._name = name
        self._status = ''

    def reload_script(self):
        file = self.path2fullpath()
#        print file
        if (file != '' and
                os.path.exists(file)):
            return self.load_script(file)
        return ''

    def onProjTreeActivate(self, e):
        self.onEditScript(e)

    def rename(self, new,  ignore_name_readonly=False):
        osfile = self.path2fullpath()
        oname = self.name
        oowndir = self.owndir()

        super(PyScript, self).rename(
            new, ignore_name_readonly=ignore_name_readonly)

        sfile = self.path2fullpath()
        mode = self.getvar("pathmode")
        if mode == 'owndir':
            nsfile = os.path.join(self.owndir(), self.name+'.py')
            if os.path.exists(sfile):
                os.rename(sfile, nsfile)
                self.set_path_pathmode(nsfile)
            if os.path.exists(sfile+'c'):
                os.remove(sfile+'c')
            if os.path.exists(sfile+'o'):
                os.remove(sfile+'o')
        else:
            nsfile = sfile
        param = {"oldname": osfile, "newname": nsfile}
        op = 'rename'

        self.load_script(self.path2fullpath())
        ifigure.events.SendFileChangedEvent(self, operation=op,
                                            param=param)

        return

    def save2(self, fid=None, olist=None):
        if self._save_link:
            return self.save2_link(fid=fid, olist=olist)
        else:
            return super(PyScript, self).save2(fid=fid, olist=olist)

    def save2_link(self, fid=None, olist=None):
        obj = PyScriptLink(obj=self)
        h2 = {"name": obj.__class__.__name__,
              "module": obj.__class__.__module__,
              "num_child": 0,
              "sname": self.name,
              "id": self.id,
              "var0": None,
              "var": collections.OrderedDict(),
              "note": collections.OrderedDict(),
              "format": 2}
#        print h2
        if fid is not None:
            pickle.dump(h2, fid)

        c = olist.count(self.id)
        if c == 0:
            data = obj.save_data2({})
            pickle.dump(data, fid)
            olist.append(self.id)
        else:
            pass

        return olist



'''
   FileHolder is a unitiliy class to support file
   in project tree

   goal of this utility is to provide

   conversion between fullpath <-> pathmode, path, ext
   provide a routine to file time stamp
'''
import os
import ifigure
from os.path import expanduser


class FileHolder(object):
    def split_ext(self, path):
        arr = path.split('.')
        base = arr[0]
        ext = '.'.join(arr[1:])
        return base, ext

    # utility to convert fullpath to path
    def set_path_pathmode(self, path,
                          modename='pathmode',
                          pathname='path',
                          extname='ext', checklist=None):
        if path == '':
            return
        pathmode, subpath = self.fullpath2path(path, checklist=checklist)
        self.setvar(modename, pathmode)
        self.setvar(pathname, subpath)
        base, ext = self.split_ext(subpath)
        self.setvar(extname, ext)
        if pathmode == 'owndir':
            if not pathname in self._items:
                self._items.append(pathname)
        else:
            if pathname in self._items:
                self._items.remove(pathname)

    def fullpath2path(self, path, checklist=None):
        '''
           script path is managed by using the
           combination of pathmode and path.
           pathmode can be one of the following
           choice
              std:  in ifigure (root = ifigure)
              wdir: in project dir (root = wdir)
              owndir: in object dir (root = self.owndir())
              usrdir: in usr_script dir
              home: in home dir (root = home dirctory)
              abs : elsewhere (root = '/')
              extfolder ; in libary path
              proj: in proj file dir 
                    (root = os.path.dirname(proj.getvar('filename'))
           path shows the relative path to the root
        '''
        import ifigure

        #### these field are not used anymore (2012.07) ####
        if self.getvar("module file") is not None:
            self.delvar("module file")
        if self.getvar("script file") is not None:
            self.delvar("script file")
        ####################################################

        # first check if owndir is used
        # if os.path.dirname(path) == self.owndir():
        #   pathmode = 'owndir'
        #   subpath = os.path.basename(path)
        #   return pathmode, subpath

        # check other possibility
        if checklist is None:
            checklist = ['owndir', 'wdir', 'std',
                         'usrpath', 'extfolder', 'proj', 'home']

        pathmode = 'abs'
        subpath = path

        def get_base(mode):
            if mode == 'owndir':
                base = self.owndir()
            elif mode == 'wdir':
                base = self.get_root_parent().getvar("wdir")
            elif mode == 'std':
                base = os.path.dirname(ifigure.__file__)
            elif mode == 'usrpath':
                base = self.get_root_parent(
                ).app.config.setting['usr_addon_dir']
                base = os.path.expanduser(base)
            elif mode == 'proj':
                base = self.get_root_parent().getvar('filename')
                if base is None:
                    return False
                base = os.path.dirname(base)
            elif mode == 'extfolder':
                base = self.get_extfolderpath()
                if base is None:
                    return False
                #base = os.path.dirname(base)
            elif mode == 'home':
                base = expanduser("~")
#                base = os.path.abspath(os.getenv('HOME'))
            else:
                return False
            return base

        def check_path_mode_match(path, base):
            c1 = os.path.commonprefix([base, path])
            if os.path.exists(c1):
                c1 = os.path.abspath(c1)  # omit finishing '/'
                if c1 == base:
                    subpath = path[len(base):]
                    return True, subpath[1:]
            return False, ''

        for mode in checklist:
            base = get_base(mode)
            if base == False:
                continue
#            print 'how about', mode
#            print base
            check, subpath = check_path_mode_match(path, base)
#            print check, subpath
            if check:
                return mode, subpath

        return 'abs', path

        '''     
        base = self.get_root_parent().getvar("wdir")
        c1 = os.path.commonprefix([base, path])
        if os.path.exists(c1):
           c1 = os.path.abspath(c1) # omit finishing '/'
           if c1== base:
              pathmode = 'wdir'
              subpath = path[len(base):]
              return pathmode, subpath[1:]

        base = os.path.dirname(ifigure.__file__)
        c1 = os.path.commonprefix([base, path])
        if os.path.exists(c1):
           c1 = os.path.abspath(c1) # omit finishing '/'
           if c1== base:
              pathmode = 'std'
              subpath = path[len(base):]
              return pathmode, subpath[1:]
 
        base = self.get_root_parent().app.config.setting['usr_addon_dir']
        base = os.path.expanduser(base)
        c1 = os.path.commonprefix([base, path])  
        if os.path.exists(c1):
           if c1 == base:
              pathmode = 'usrpath'
              subpath = path[len(base):]
              return pathmode, subpath[1:]

        base = self.get_root_parent().getvar('filename')            
        base = os.path.dirname(base)
        c1 = os.path.commonprefix([base, path])  
        if os.path.exists(c1):
           if c1 == base:
              pathmode = 'proj'
              subpath = path[len(base):]
              return pathmode, subpath[1:]

        base = os.path.abspath(os.getenv('HOME'))
        c1 = os.path.commonprefix([base, path])  
        if os.path.exists(c1):
           if c1 == base:
              pathmode = 'home'
              subpath = path[len(base):]
              return pathmode, subpath[1:]
        '''

    def path2fullpath(self, modename='pathmode', pathname='path'):
        #        print 'entering path2fullpath'
        from os.path import expanduser
        mode = 'std'
        if self.hasvar(modename):
            mode = self.getvar(modename)
        else:
            return ''
        if mode == 'std':
            base = os.path.dirname(ifigure.__file__)
        elif mode == 'wdir':
            base = self.get_root_parent().getvar('wdir')
        elif mode == 'home':
            base = os.getenv('HOME')
        elif mode == 'owndir':
            base = self.owndir()
        elif mode == 'extfolder':
            base = self.get_extfolderpath()
            if base is None:
                base = '/'
            base = expanduser(base)
        elif mode == 'usrpath':
            base = self.get_root_parent().app.config.setting['usr_addon_dir']
            base = os.path.expanduser(base)
#           if not os.path.isabs(base):
#              base = os.path.join(os.path.abspath(os.getenv('HOME')), base)
        elif mode == 'proj':
            base = self.get_root_parent().getvar('filename')
            base = os.path.dirname(base)
        else:
            base = '/'
#        print mode, path

        from ifigure.utils.cbook import isstringlike
        if (self.hasvar(pathname)):
            # this check is to recovery from a bug
            # once introduced before.
            if not isstringlike(self.getvar(pathname)):
                self.delvar(pathname)
                return ''
            return os.path.join(base, self.getvar(pathname))
        else:
            return ''

    def get_mtime(self, modename='pathmode', pathname='path'):
        return os.path.getmtime(self.path2fullpath(modename, pathname))

    def store_mtime(self, mname='mtime', modename='pathmode', pathname='path'):
        mtime = self.get_mtime(modename=modename, pathname=pathname)
        self.setvar(mname, mtime)

    def isExternalFileNewer(self, mname='mtime', modename='pathmode', pathname='path'):
        if self.isInternalFile():
            return False
        cmtime = os.path.getmtime(self.path2fullpath(
            modename='pathmode', pathname='path'))
        if self.getvar(mname) is None:
            return False
        return (cmtime > self.getvar(mname))

    def isInternalFile(self):
        if (self.getvar('file_pathmode') == 'wdir' or
                self.getvar('file_pathmode') == 'owndir'):
            return True
        return False



from ifigure.interactive import figure
from ifigure.widgets.book_viewer import BookViewer
import numpy as np
import wx


class BookViewerGui(BookViewer):
    def __init__(self, *args, **kargs):
        BookViewer.__init__(self, *args, **kargs)
        extra_menu = wx.Menu()
        self.menuBar.Insert(self.menuBar.GetMenuCount()-1,
                            extra_menu, "ExtraMenu")
        self.add_menu(extra_menu, wx.ID_ANY,
                      "Plot Something...",
                      "plot sine wave (this is string for status bar)",
                      self.onPlotSomething)

    def onPlotSomething(self, evt):
        t = np.arange(1000.)/999.
        f = 10.
        self.plot(t, np.sin(2*3.14*f*t))


if not obj.get_parent().has_child('bookgui_book'):
    v = figure()
    v.book.rename('bookgui_book')
    v.book.set_keep_data_in_tree(True)
    v.book.Close()

obj.get_parent().bookgui_book.Open(BookViewerGui)



"""Functions for getting memory usage of Windows processes."""

__all__ = ['get_current_process', 'get_memory_info', 'get_memory_usage']

import ctypes
from ctypes import wintypes

GetCurrentProcess = ctypes.windll.kernel32.GetCurrentProcess
GetCurrentProcess.argtypes = []
GetCurrentProcess.restype = wintypes.HANDLE

SIZE_T = ctypes.c_size_t


class PROCESS_MEMORY_COUNTERS_EX(ctypes.Structure):
    _fields_ = [
        ('cb', wintypes.DWORD),
        ('PageFaultCount', wintypes.DWORD),
        ('PeakWorkingSetSize', SIZE_T),
        ('WorkingSetSize', SIZE_T),
        ('QuotaPeakPagedPoolUsage', SIZE_T),
        ('QuotaPagedPoolUsage', SIZE_T),
        ('QuotaPeakNonPagedPoolUsage', SIZE_T),
        ('QuotaNonPagedPoolUsage', SIZE_T),
        ('PagefileUsage', SIZE_T),
        ('PeakPagefileUsage', SIZE_T),
        ('PrivateUsage', SIZE_T),
    ]


GetProcessMemoryInfo = ctypes.windll.psapi.GetProcessMemoryInfo
GetProcessMemoryInfo.argtypes = [
    wintypes.HANDLE,
    ctypes.POINTER(PROCESS_MEMORY_COUNTERS_EX),
    wintypes.DWORD,
]
GetProcessMemoryInfo.restype = wintypes.BOOL


def get_current_process():
    """Return handle to current process."""
    return GetCurrentProcess()


def get_memory_info(process=None):
    """Return Win32 process memory counters structure as a dict."""
    if process is None:
        process = get_current_process()
    counters = PROCESS_MEMORY_COUNTERS_EX()
    ret = GetProcessMemoryInfo(process, ctypes.byref(counters),
                               ctypes.sizeof(counters))
    if not ret:
        raise ctypes.WinError()
    info = dict((name, getattr(counters, name))
                for name, _ in counters._fields_)
    return info


def get_memory_usage(process=None):
    """Return this process's memory usage in bytes."""
    info = get_memory_info(process=process)
    return info['PrivateUsage']


if __name__ == '__main__':
    import pprint
    pprint.pprint(get_memory_info())



from ifigure.mto.fig_contour import FigContour
from ifigure.mto.triangle_plots import TrianglePlots
from ifigure.utils.cbook import ProcessKeywords
from matplotlib.tri import Triangulation
import numpy as np
import ifigure.events


class FigTricontour(FigContour, TrianglePlots):
    '''
        tricontour : contour plot using triangulation
        tricontour(x, y, z, n)  
        tricontour(x, y, z, v)  
        tricontour(tri, x, y, z, v)  
        tricontour(tri, x, y, z, n)  

        n: number of levels
        v: a list of contour levels
    '''
    default_rasterized = True

    def __new__(self, *args, **kywds):
        if len(args) == 5:
            tri = args[0]
            args = args[1:]
        else:
            tri = None
        kywds['use_tri'] = True
        mask = kywds.pop('mask', None)
        obj = FigContour.__new__(self, *args, **kywds)
        obj.setp('mask', mask)
        obj.setvar('mask', mask)
        return obj

    def __init__(self, *args,  **kywds):
        if len(args) == 4:
            tri = args[0]
            args = args[1:]
        else:
            tri = None
        kywds['use_tri'] = True
        FigContour.__init__(self, *args, **kywds)
        TrianglePlots.__init__(self)
        self._tri = tri

    @classmethod
    def isTripContour(self):
        return True

    @classmethod
    def get_namebase(self):
        return 'tricontour'

#    @classmethod
#    def attr_in_file(self):
#        return  (["alpha", "shading", "mask"] +
#                super(FigImage, self).attr_in_file())
#    @classmethod
#    def property_in_palette(self):
#        return ["cmap","tripcolor_shading", "alpha_2"]

    def onResize(self, evt):
        self.set_bmp_update(False)

        # self.del_artist(delall=True)
        # self.delp('loaded_property')
        #self.setp('use_var', False)
        # if hasattr(self, "_data_extent"):
        #   self._data_extent = None
        # self.generate_artist()

    def handle_axes_change(self, evt=None):
        return super(FigContour, self).handle_axes_change(evt)

    def set_alpha(self, value, a):
        a.set_alpha(value)
        cax = self.get_caxisparam()
        cax.set_crangeparam_to_artist(a)
        self.setp('alpha', value)
        self.set_bmp_update(False)

    def get_alpha(self, a):
        return a.get_alpha()



from __future__ import print_function
import time
import weakref
import threading
from six.moves import queue as Queue

queues = []


def get_thread():
    queues.append(Queue.Queue())
    return EventDrivenThread(queues[-1]), queues[-1]


def send_event(event, t=None):
    if t is None:
        name = threading.current_thread().name
    else:
        name = t.name
    for q in queues:
        q.put(name+'_'+event)


def remove_queue(q):
    if q in queues:
        queues.remove(q)


class EventDrivenThread(threading.Thread):
    def __init__(self, queue, *args, **kargs):
        super(EventDrivenThread, self).__init__(*args, **kargs)
        self.queue = queue
        self.events = {}
        self.start_event = 'start_action'
        self.finish_event = 'finish_action'
        self.quit_event = 'quit_thread'

    def bind(self, event, method, t=None):
        ''' 
        bind an evnet from thread t to method 
        '''
        if t is None:
            t = self
        ename = t.name + '_' + event
        if ename in self.events:
            self.events[ename].append(method)
        else:
            self.events[ename] = [method]

    def unbind(self, event, method, t=None):
        if t is None:
            t = self
        ename = t.name + '_' + event
        if ename in self.events:
            if method in self.events[ename]:
                self.events[ename].remove(method)

    def bind_init(self, method):
        self.bind(self.start_event,  method)

    def bind_quit(self):
        self.bind(self.quit_event, self.onQuitThread)

    def run(self, *args, **kagrs):
        send_event(self.start_event)
        self.flag = True
        while (self.flag):
            if self.queue.empty():
                time.sleep(0.1)
            else:
                event = self.queue.get(False)
                print(event)
                print(self.events)
                if event in self.events:
                    print(('calling method for ', event))
                    for m in self.events[event]:
                        m(event)
        remove_queue(self.queue)
        send_event(self.finish_event)
        self.queue = None
        self.events = None
        print(('Exiting...', self.name))

    def onQuitThread(self, e):
        self.flag = False
